"""
Tabs
====

See: https://www.carbondesignsystem.com/components/tabs/usage/

Use tabs to allow users to navigate easily between views within the same
context.

Overview
--------

Tabs are used to quickly navigate between views within the same context.
""" # pylint:disable=line-too-long
# pylint:disable=too-many-lines

from .base import Node

class Tabs(Node):
    """Tabs component.
    """
    WANT_CHILDREN = True
    "Template Tag needs closing end tag."
    SLOTS = ('header',)
    "Named children."
    NODE_PROPS = ('container',)
    "Extended Template Tag arguments."

    def prepare(self, values, context):
        """Prepare values for rendering the templates.
        """
        if self.eval(self.kwargs.get('container'), context):
            values['class'].append('bx--tabs--container')


    def render_default(self, values, context):
        """Output html of the component.
        """
        template = """
<div data-tabs class="bx--tabs {class}" {props}>
  <div class="bx--tabs-trigger" tabindex="0">
    <a href="javascript:void(0)" class="bx--tabs-trigger-text" tabindex="-1"></a>
    <svg focusable="false" preserveAspectRatio="xMidYMid meet"
        xmlns="http://www.w3.org/2000/svg" fill="currentColor" width="16"
        height="16" viewBox="0 0 16 16" aria-hidden="true">
      <path d="M8 11L3 6 3.7 5.3 8 9.6 12.3 5.3 13 6z"></path>
    </svg>
  </div>
  <ul class="bx--tabs__nav bx--tabs__nav--hidden" role="tablist">
    {slot_header}
  </ul>
</div>
<div class="bx--tab-content">
  {child}
</div>
"""
        return self.format(template, values, context)


class TabItem(Node):
    """Tabs item component.
    """
    WANT_CHILDREN = True
    "Template Tag needs closing end tag."
    NODE_PROPS = ('active', 'target', 'disabled')
    "Extended Template Tag arguments."
    CLASS_AND_PROPS = ('tab',)
    "Prepare xxx_class and xxx_props values."

    def prepare(self, values, context):
        """Prepare values for rendering the templates.
        """
        active = self.eval(self.kwargs.get('active'), context)
        if active:
            values['tab_class'].append('bx--tabs__nav-item--selected')
            values['tab_props'].append(('aria-selected', 'true'))

        target = self.eval(self.kwargs.get('target'), context)
        values['target'] = target or ''
        if target:
            values['props'].append(('id', f'tab-link-{target}'))

        if self.eval(self.kwargs.get('disabled'), context):
            values['tab_class'].append('bx--tabs__nav-item--disabled')
            values['tab_props'].append(('aria-disabled', 'true'))


    def render_default(self, values, context):
        """Output html of the component.
        """
        template = """
<li class="bx--tabs__nav-item {tab_class}" data-target="#{target}"
    role="tab" {tab_props}>
  <a tabindex="0" class="bx--tabs__nav-link {class}" href="javascript:void(0)"
      role="tab" aria-controls="{target}" {props}>
    {child}
  </a>
</li>
"""
        return self.format(template, values)


class TabContent(Node):
    """Tabs content component.
    """
    WANT_CHILDREN = True
    "Template Tag needs closing end tag."
    NODE_PROPS = ('id', 'active', 'target')
    "Extended Template Tag arguments."

    def prepare(self, values, context):
        """Prepare values for rendering the templates.
        """
        active = self.eval(self.kwargs.get('active'), context)
        if active:
            values['props'].append(('aria-hidden', 'false'))
        else:
            values['props'].append(('aria-hidden', 'true'))
            values['props'].append(('hidden', True))

        values['target'] = self.eval(self.kwargs.get('target'), context)


    def render_default(self, values, context):
        """Output html of the component.
        """
        template = """
<div id="{id}" role="tabpanel" aria-labelledby="tab-link-{id}" {props}>
  <div>{child}</div>
</div>
"""
        return self.format(template, values)


components = {
    'Tabs': Tabs,
    'TabItem': TabItem,
    'TabContent': TabContent,
}



"""
List
====

See: https://www.carbondesignsystem.com/components/list/usage/

Lists are vertical groupings of related content. List items begin with either
a number or a bullet.

Overview
--------

Lists consist of related content grouped together and organized vertically.

Use bulleted lists when you don’t need to convey a specific order for list
items.

Use numbered lists when you need to convey a priority, hierarchy, or sequence
between list items.
""" # pylint:disable=line-too-long
# pylint:disable=too-many-lines

from .base import Node

class List(Node):
    """List component.
    """
    WANT_CHILDREN = True
    "Template Tag needs closing end tag."
    NODE_PROPS = ('native',)
    "Extended Template Tag arguments."
    DEFAULT_TAG = 'ul'
    "Rendered HTML tag."

    def prepare(self, values, context):
        """Prepare values for rendering the templates.
        """
        if values['tag'] == 'ul':
            values['class'].append('bx--list--unordered')
        elif self.eval(self.kwargs.get('native'), context):
            values['class'].append('bx--list--ordered--native')
        else:
            values['class'].append('bx--list--ordered')

        if context.get('list_nested'):
            values['class'].append('bx--list--nested')
        context['list_nested'] = True


    def render_default(self, values, context):
        """Output html of the component.
        """
        template = """
<{tag} class="{class}" {props}>
  {child}
</{tag}>
"""
        return self.format(template, values)


class ListItem(Node):
    """List item component.
    """
    WANT_CHILDREN = True
    "Template Tag needs closing end tag."

    def render_default(self, values, context):
        """Output html of the component.
        """
        template = """
<li class="bx--list__item {class}" {props}>
  {child}
</li>
"""
        return self.format(template, values)


components = {
    'List': List,
    'Li': ListItem,
}



"""
Multi Select
============

See: https://the-carbon-components.netlify.app/?nav=multi-select

""" # pylint:disable=line-too-long
# pylint:disable=too-many-lines

from django.utils.translation import gettext as _
#-
from .base import FormNode

class MultiSelect(FormNode):
    """Multi Select component.
    """
    MODES = ('default', 'inline', 'filterable')
    "Available variants."
    NODE_PROPS = ('light', 'expanded')
    "Extended Template Tag arguments."
    CLASS_AND_PROPS = ('help', 'list', 'trigger')
    "Prepare xxx_class and xxx_props values."

    def prepare(self, values, context):
        """Prepare values for rendering the templates.
        """
        values['txt_open'] = _("Open menu")
        values['txt_close'] = _("Close menu")
        values['txt_multi'] = _("Multi select options")
        values['txt_clear'] = _("Clear all selected items")
        values['txt_clear_num'] = _("Clear selection")
        values['txt_filter'] = _("Filter...")

        if self.bound_value is None:
            self.bound_value = []
        elif isinstance(self.bound_value, str):
            self.bound_value = [self.bound_value]
        count = len(self.bound_value)
        values['selected_count'] = count

        if self.eval(self.kwargs.get('light'), context):
            values['list_class'].append('bx--list-box--light')

        if self.eval(self.kwargs.get('expanded'), context):
            values['list_class'].append('bx--list-box--expanded')
            values['trigger_props'].append(('aria-expanded', 'true'))
            values['txt_menu'] = values['txt_close']
        else:
            values['trigger_props'].append(('aria-expanded', 'false'))
            values['txt_menu'] = values['txt_open']

        if count:
            values['list_class'].append('bx--multi-select--selected')


    def render_default(self, values, context):
        """Output html of the component.
        """
        template = """
<div class="bx--form-item">
  <div class="bx--list-box__wrapper">
    <label class="bx--label">
      {label}
    </label>
    <div class="bx--multi-select bx--list-box {list_class}">
      <div role="button" class="bx--list-box__field {trigger_class}" tabindex="0"
          aria-label="{txt_menu}" aria-haspopup="true" {trigger_props}>
        {tmpl_icon_clear}
        <span class="bx--list-box__label">{txt_multi}</span>
        {tmpl_icon_menu}
      </div>
      <fieldset class="bx--list-box__menu" role="listbox">
        <legend class="bx--assistive-text">
          {label}
        </legend>
        {tmpl_items}
      </fieldset>
    </div>
    {tmpl_help}
  </div>
</div>
"""
        return self.format(template, values, context)


    def render_filterable(self, values, context):
        """Output html of the component.
        """
        template = """
<div class="bx--form-item">
  <div class="bx--list-box__wrapper">
    <label class="bx--label">
      {label}
    </label>
    <div class="bx--multi-select bx--list-box bx--combo-box bx--multi-select--filterable {list_class}">
      <div role="button" class="bx--list-box__field {trigger_class}" tabindex="0"
          aria-label="{txt_menu}" aria-haspopup="true" {trigger_props}>
        {tmpl_icon_clear}
        <input class="bx--text-input" placeholder="{txt_filter}">
        {tmpl_icon_menu}
      </div>
      <fieldset class="bx--list-box__menu" role="listbox">
        <legend class="bx--assistive-text">
          {label}
        </legend>
        {tmpl_items}
      </fieldset>
    </div>
    {tmpl_help}
  </div>
</div>
"""
        return self.format(template, values, context)


    def render_inline(self, values, context):
        """Output html of the component.
        """
        template = """
<div class="bx--form-item">
  <div class="bx--list-box__wrapper bx--list-box__wrapper--inline">
    <label class="bx--label">
      {label}
    </label>
    <div class="bx--multi-select bx--list-box bx--list-box--inline {list_class}">
      <div role="button" class="bx--list-box__field {trigger_class}" tabindex="0"
          aria-label="{txt_menu}" aria-haspopup="true" {trigger_props}>
        {tmpl_icon_clear}
        <span class="bx--list-box__label">{txt_multi}</span>
        {tmpl_icon_menu}
      </div>
      <fieldset class="bx--list-box__menu" role="listbox">
        <legend class="bx--assistive-text">
          {label}
        </legend>
        {tmpl_items}
      </fieldset>
    </div>
    {tmpl_help}
  </div>
</div>
"""
        return self.format(template, values, context)


    def render_tmpl_icon_clear(self, values, context):
        """Dynamically render a part of the component's template.
        """
        if not values['selected_count']:
            return ''
        template = """
<div role="button"
    class="bx--list-box__selection bx--list-box__selection--multi bx--tag--filter"
    tabindex="0" title="{txt_clear}">
  {selected_count}
  <svg focusable="false" preserveAspectRatio="xMidYMid meet"
      xmlns="http://www.w3.org/2000/svg" fill="currentColor"
      aria-label="{txt_clear_num}" width="16" height="16"
      viewBox="0 0 32 32" role="img">
    <path d="M24 9.4L22.6 8 16 14.6 9.4 8 8 9.4 14.6 16 8 22.6 9.4 24 16 17.4 22.6 24 24 22.6 17.4 16 24 9.4z"></path>
  </svg>
</div>
"""
        return self.format(template, values)


    def render_tmpl_icon_menu(self, values, context):
        """Dynamically render a part of the component's template.
        """
        if self.eval(self.kwargs.get('expanded'), context):
            template = """
<div class="bx--list-box__menu-icon">
  <svg focusable="false" preserveAspectRatio="xMidYMid meet"
      xmlns="http://www.w3.org/2000/svg" fill="currentColor"
      aria-label="{txt_close}" width="16" height="16" viewBox="0 0 16 16"
      role="img">
    <path d="M8 5L13 10 12.3 10.7 8 6.4 3.7 10.7 3 10z"></path>
  </svg>
</div>
"""
        else:
            template = """
<div class="bx--list-box__menu-icon">
  <svg focusable="false" preserveAspectRatio="xMidYMid meet"
      xmlns="http://www.w3.org/2000/svg" fill="currentColor"
      aria-label="{txt_open}" width="16" height="16" viewBox="0 0 16 16"
      role="img">
    <path d="M8 11L3 6 3.7 5.3 8 9.6 12.3 5.3 13 6z"></path>
  </svg>
</div>
"""
        return self.format(template, values)


    def render_tmpl_items(self, values, context):
        """Dynamically render a part of the component's template.
        """
        template = """
<div class="bx--list-box__menu-item">
  <div class="bx--list-box__menu-item__option">
    <div class="bx--form-item bx--checkbox-wrapper">
      <label title="{child}" class="bx--checkbox-label">
        <input type="checkbox" name="{name}" readonly class="bx--checkbox"
            id="{id}" value="{value}" {props}>
        <span class="bx--checkbox-appearance"></span>
        <span class="bx--checkbox-label-text">
          {child}
        </span>
      </label>
    </div>
  </div>
</div>
"""
        items = []
        for ii, (_, val, txt) in enumerate(self.choices()):
            options = {
                'id': '%s-%s' % (values['id'], ii + 1),
                'value': val,
                'child': txt,
                'name': self.bound_field.name,
            }
            props = []
            if val in self.bound_value:
                props.append('checked')
            options['props'] = ' '.join(props)
            items.append(self.format(template, options))

        return '\n'.join(items)


components = {
    'MultiSelect': MultiSelect,
}



"""Django application configuration for carbondesign.
"""
from django.apps import AppConfig


class CarbonDesignConfig(AppConfig):
    """Django application configuration for carbondesign.
    """
    name = 'carbondesign'



"""Base classes for carbondesign Template Tags.
"""
import html
import logging
import re
from typing import Sequence
from uuid import uuid4
#-
from django import template
from django.template.base import TextNode # pylint:disable=unused-import
from django.utils.html import strip_tags
from django.utils.translation import gettext as _
from lxml import etree
#-
from .base_widgets import CustomCheckboxInput, CustomTextarea, CustomTextInput

_logger = logging.getLogger(__name__)

SLOT_NAME_PATTERN = re.compile(r'{(tmpl|slot)_(\w+)}')
TMPL_MULTI_PATTERN = re.compile(r'(?P<name>\w+)_(?P<index>\d+)$')


def var_eval(value, context):
    """Evaluate argument passed to our Template Tag.
    """
    if isinstance(value, template.Variable):
        return value.resolve(context)
    return value


def modify_svg(xml, props):
    """Modify xml attributes of an svg image.
    """
    # pylint:disable=c-extension-no-member
    root = etree.fromstring(xml)
    for attr, value in props.items():
        if attr == 'style':
            value = ';'.join('%s:%s' % (x, y) for x, y in value.items())
        root.attrib[attr] = value
    return etree.tostring(root).decode()


def clean_attr_value(value):
    """Strip html of its tags and prepare it to be used as html attribute value.
    """
    value = strip_tags(value).strip().replace('\n', ' ')
    return html.escape(re.sub(r'\s\s+', ' ', value))


class IgnoreMissing(dict):
    """String formatting but the variables are optional.
    """
    def __missing__(self, key):
        """Replacement string for missing variables.
        """
        return '{' + key + '}'


class DummyNodeList:
    """Fake Django template node.
    """
    def __init__(self, text):
        self.text = text


    def get_nodes_by_type(self, type_):
        """Dummy method that returns empty.
        """
        return []


    def render(self, context):
        """Return the passed in string.
        """
        return self.text


class Slot(template.Node):
    """Implements slots.

    It's very simple, so don't inherit from Node class below.
    """
    WANT_CHILDREN = True
    "Template Tag needs closing end tag."

    name = None

    def __init__(self, *args, **kwargs):
        # First argument sent by templatetag is children nodes.
        self.nodelist = args[0]
        # Second required argument is the slot's name.
        self.name = args[1]
        self.kwargs = kwargs


    def classList(self, context):
        """Get class argument as a list.
        """
        return var_eval(self.kwargs.get('class', ''), context).split()


    def label(self, context):
        """Get label argument.
        """
        return var_eval(self.kwargs.get('label', ''), context)


    def props(self, context):
        """Filter arguments passed to Slot excluding the PROPS.
        """
        props = []
        for key, val in self.kwargs.items():
            if key in ('class', 'label'):
                continue
            props.append((key, var_eval(val, context)))
        # Ignore properties with falsy value except empty string.
        return [x for x in props if bool(x[1]) or x[1] == '']


    def render(self, context):
        """Render the Slot as html.
        """
        return self.nodelist.render(context)


class Node(template.Node):
    """Base class for all carbondesign tags.
    """
    WANT_CHILDREN = False
    "Template Tag needs closing end tag."
    SLOTS = ()
    "Named children."
    MODES = ()
    "Available variants."
    BASE_NODE_PROPS = ('mode', 'tag', 'class', 'label')
    "Base Template Tag arguments."
    NODE_PROPS = ()
    "Extended Template Tag arguments."
    REQUIRED_PROPS = ()
    "Will raise Exception if not set."
    DEFAULT_TAG = 'div'
    "Rendered HTML tag."
    CLASS_AND_PROPS = ()
    "Prepare xxx_class and xxx_props values."

    # Parent Tags can set the arguments of their children Tags, in effect
    # changing their appearance.
    CATCH_PROPS = ()

    context = None
    mode = None
    _id = None

    def __init__(self, *args, **kwargs):
        if self.WANT_CHILDREN:
            self.args = args[1:]

            self.nodelist = args[0]
            slots = self.nodelist.get_nodes_by_type(Slot)
            for slot in slots:
                self.nodelist.remove(slot)
            self.slots = {s.name: s for s in slots}
        else:
            self.args = args
            self.slots = {}

        self.kwargs = kwargs


    def render(self, context):
        """Render the Template Tag as html.
        """
        # Create local context.
        context.push({})
        # Parent Tags can set the arguments of their children Tags.
        # You can also set them to children Tags in specific slot.
        myslot = context.get('slot')
        for drop in self.CATCH_PROPS:
            if drop in context:
                self.kwargs.update(context[drop])
            if myslot and f'{myslot}_{drop}' in context:
                self.kwargs.update(context[f'{myslot}_{drop}'])

        for prop in self.REQUIRED_PROPS:
            if prop not in self.kwargs:
                raise template.exceptions.TemplateSyntaxError(
                        _("Argument {prop} for template tag is required.")\
                        .format(prop=prop))

        values = {}

        self.before_prepare(values, context)
        self.prepare(values, context)

        if self.WANT_CHILDREN:
            values['child'] = self.nodelist.render(context)
        else:
            values['child'] = ''

        self.after_prepare(values, context)

        method = getattr(self, 'render_%s' % self.mode, None)
        if not method:
            raise NotImplementedError("Method is missing: render_%s" %\
                    self.mode)

        result = method(values, context) # pylint:disable=not-callable
        # Destroy local context.
        context.pop()
        return result


    def default_id(self):
        """Get unique id.
        """
        return 'node-' + uuid4().hex


    def props(self, context):
        """Filter arguments passed to Template Tag excluding the PROPS.
        """
        props = [(key, val) for key, val in self.kwargs.items()\
                if key not in self.BASE_NODE_PROPS\
                and key not in self.NODE_PROPS \
                and key not in (f'{a}_class' for a in self.CLASS_AND_PROPS)]
        props = [(key, var_eval(val, context)) for key, val in props]
        # Ignore properties with falsy value except empty string.
        return [x for x in props if bool(x[1]) or x[1] == '']


    def tmpl(self, name, values, context, slots):
        """Render individual templates.

        We don't return values like we do in javascript, not needed.
        """
        slot_name = f'tmpl_{name}'
        if slot_name in slots:
            return

        method = getattr(self, f'render_tmpl_{name}')
        slots[slot_name] = method(values, context)


    def slot(self, name, values, context, slots):
        """Render individual slots.
        """
        # We don't return values like we do in javascript, not needed.
        slot_name = f'slot_{name}'
        if slot_name in slots:
            return

        slot = self.slots.get(name)
        if slot:
            method = getattr(self, f'render_slot_{name}', None)
            if method:
                # Process values set in self.before_prepare_slots()
                self.after_prepare_class_props(name, values, context)

                # Create slot local context.
                context.push({'slot': name})
                slots[slot_name] = method(
                        {
                            # Use context to tell slot's children in which slot
                            # they are, for example when catching props from
                            # parent tags.
                            # It is set here to mimic javascript.
                            'child': slot.render(context),
                            'class': values[name + '_class'],
                            'props': values[name + '_props'],
                            'id': values['id'],
                            'label': slot.label(context),
                        },
                        context)
                # Destroy slot local context.
                context.pop()
            else:
                slots[slot_name] = slot.render(context)
        else:
            slots[slot_name] = ''


    def before_prepare(self, values, context):
        """Initialize the values meant for rendering templates.
        """
        self.mode = self.kwargs.get('mode', None)
        if self.MODES:
            if not self.mode:
                self.mode = self.MODES[0]
            elif self.mode not in self.MODES:
                raise NotImplementedError("Mode %s is not allowed." %\
                        self.mode)
        else:
            self.mode = 'default'

        values['mode'] = self.mode
        values['tag'] = var_eval(self.kwargs.get('tag', self.DEFAULT_TAG),
                context)
        values['props'] = self.props(context)
        values['class'] = var_eval(self.kwargs.get('class', ''), context)\
                .split()
        self._id = var_eval(self.kwargs.get('id', self.default_id()), context)
        values['id'] = self._id
        values['label'] = var_eval(self.kwargs.get('label', ''), context)
        for name in self.CLASS_AND_PROPS:
            if name in self.SLOTS:
                continue
            self.before_prepare_class_props(name, values, context)

        self.before_prepare_slots(values, context)


    def before_prepare_class_props(self, name, values, context):
        """Initialize variables for storing html classes and attributes.
        """
        values[f'{name}_props'] = []
        values[f'{name}_class'] = var_eval(
                self.kwargs.get(f'{name}_class', ''), context)\
                .split()


    def before_prepare_slots(self, values, context):
        """Initialize the values meant for rendering slot templates.
        """
        for name in self.SLOTS:
            slot = self.slots.get(name)
            if slot:
                values[f'{name}_class'] = slot.classList(context)
                values[f'{name}_props'] = slot.props(context)
            else:
                values[f'{name}_class'] = []
                values[f'{name}_props'] = []


    def after_prepare(self, values, context):
        """Simplifying values meant for rendering templates.
        """
        values['props_raw'] = self.prune_attributes(values['props'])
        values['props'] = self.join_attributes(self.prune_attributes(
                values['props']))
        values['class'] = ' '.join(values['class'])

        for name in self.CLASS_AND_PROPS:
            if name in self.SLOTS:
                continue
            self.after_prepare_class_props(name, values, context)


    def after_prepare_class_props(self, name, values, context):
        """Join html classes and html attributes into single strings.
        """
        values[f'{name}_props'] = self.join_attributes(
                self.prune_attributes(values[f'{name}_props']))
        values[f'{name}_class'] = ' '.join(values[f'{name}_class'])


    def prune_attributes(self, attrs):
        """Cleanup duplicate html attributes.
        """
        added_props = set()

        def _clean_attributes():
            for prop in reversed(attrs):
                prop_name = prop[0]
                if prop_name in added_props:
                    continue
                added_props.add(prop_name)
                yield prop

        return reversed(list(_clean_attributes()))


    def join_attributes(self, attrs):
        """Format html attributes.
        """
        result = []
        for att, val in attrs:
            if val is None or val is False:
                continue
            if val is True:
                result.append(att)
            else:
                val = html.escape(str(val))
                result.append(f'{att}="{val}"')
        return ' '.join(result)


    def prepare(self, values, context):
        """Prepare values for rendering the templates.
        """


    def eval(self, value, context):
        """Evaluate argument passed to our Template Tag.
        """
        return var_eval(value, context)


    def format(self, tpl, values, context=None):
        """Apply the prepared values to the templates.
        """
        if context:
            # Assume the caller want to tell us there is sub-templates.
            slots = {}
            for typ, nam in SLOT_NAME_PATTERN.findall(tpl):
                method = getattr(self, typ)
                method(nam, values, context, slots)
            if slots:
                tpl = tpl.format_map(IgnoreMissing(slots))

        try:
            return tpl.format_map(IgnoreMissing(values))
        except ValueError as e:
            raise e


    def set_child_props(self, context, name, slot=None, **kwargs):
        """Use context to set arguments for all children Template Tags.
        """
        if slot:
            name = f'{slot}_{name}'
        context.setdefault(name, {})
        context[name].update(kwargs)


class FormNode(Node):
    """Base class for form field tags.

    The first argument to the tag is the Django form field.
    """
    BASE_NODE_PROPS = ('widget', 'id', 'hidden', 'disabled',
            *Node.BASE_NODE_PROPS)
    "Base Template Tag arguments."
    CLASS_AND_PROPS = ('label', 'help')
    "Prepare xxx_class and xxx_props values."

    bound_field = None
    bound_value = None
    _choices = None

    def default_id(self):
        """Get Django form field html id.
        """
        return self.bound_field.id_for_label


    def label(self):
        """Get Django form field label.
        """
        return self.bound_field.label


    def before_prepare(self, values, context):
        """Initialize the values meant for rendering templates.
        """
        self.bound_field = self.args[0].resolve(context)
        self.bound_value = self.bound_field.value()
        super().before_prepare(values, context)
        if not values['label']:
            values['label'] = self.label()


    def prepare_element_props(self, props, context):
        """Prepare html attributes for rendering the form element.
        """


    def choices(self):
        """Get Django form field choices.
        """
        if self._choices is not None:
            return self._choices
        self._choices = []
        try:
            # We assume self.bound_field is set, but
            # self.bound_field.field.choices may not be.
            choices = self.bound_field.field.choices
        except AttributeError:
            # Assume we are dealing with BooleanField
            default_truthy = self.bound_value or 'on'
            label = self.bound_field.label
            choices = (('', label), (default_truthy, label))

        for option_value, option_label in choices:
            if option_value is None:
                option_value = ''

            if isinstance(option_label, (list, tuple)):
                group_name = option_value
                subchoices = option_label
            else:
                group_name = None
                subchoices = [(option_value, option_label)]

            for subvalue, sublabel in subchoices:
                self._choices.append((group_name, subvalue, sublabel))
        return self._choices


    def render_tmpl_element(self, values, context):
        """Render django form field.
        """
        bound_field = self.bound_field

        attrs = dict(values['props_raw'])
        attrs['id'] = self._id
        attrs['class'] = bound_field.field.widget.attrs.get('class', '').split()
        if bound_field.help_text or 'help' in self.slots:
            attrs['aria-controls'] = 'hint-' + values['id']
            attrs['aria-describedby'] = 'hint-' + values['id']

        self.prepare_element_props(attrs, context)
        attrs['class'] = ' '.join(attrs['class'])

        if var_eval(self.kwargs.get('disabled'), context):
            attrs['disabled'] = True

        if var_eval(self.kwargs.get('hidden'), context):
            return bound_field.as_hidden(attrs=attrs)

        widget = self.eval(self.kwargs.get('widget'), context)
        if widget == 'input':
            widget = CustomTextInput(attrs=attrs)
        elif widget == 'textarea':
            widget = CustomTextarea(attrs=attrs)
        else:
            return bound_field.as_widget(attrs=attrs)
        return bound_field.as_widget(widget=widget, attrs=attrs)


    def render_tmpl_label(self, values, context):
        """Dynamically render a part of the component's template.
        """
        if not values['label']:
            return ''
        tmpl = """
<label for="{id}" class="bx--label {label_class}" {label_props}>
  {label}
</label>
"""
        return self.format(tmpl, values)


    def render_tmpl_help(self, values, context):
        """Dynamically render a part of the component's template.
        """
        if self.bound_field.help_text:
            tmpl = """
<div id="hint-{id}" class="bx--form__helper-text {class}" {props}>
  {child}
</div>
"""
            help_values = {
                'child': self.bound_field.help_text,
                'class': values['help_class'],
                'props': values['help_props'],
                'id': values['id'],
            }
            return tmpl.format(**help_values)
        return ''


    def render_tmpl_errors(self, values, context):
        """Dynamically render a part of the component's template.
        """
        tmpl_t = '<div class="bx--form-requirement__title">{title}.</div>'
        tmpl_s = '<p class="bx--form-requirement__supplement">{supplement}</p>'

        items = []
        for error in self.bound_field.errors:
            title, supplement = error.split('.', 1)
            supplement = supplement.strip()
            items.append(tmpl_t.format(title=title))
            if supplement:
                items.append(tmpl_s.format(supplement=supplement.strip()))
        return '\n'.join(items)


class FormNodes(Node):
    """Base class for tags with multiple form fields.

    The arguments to the tag are all Django form fields.
    """
    BASE_NODE_PROPS = ('widget', 'hidden', 'disabled', *Node.BASE_NODE_PROPS)
    "Base Template Tag arguments."

    bound_fields = None

    def before_prepare(self, values, context):
        """Initialize the values meant for rendering templates.
        """
        self.bound_fields = [x.resolve(context) for x in self.args]
        for ii, field in enumerate(self.bound_fields):
            values[f'id_{ii}'] = field.id_for_label
            values[f'label_{ii}'] = field.label
        super().before_prepare(values, context)


    def prepare_element_props(self, props, context, bound_field):
        """Prepare html attributes for rendering the form element.
        """


    def tmpl(self, name, values, context, slots):
        """Render individual templates.

        We don't return values like we do in javascript, not needed.
        """
        if isinstance(name, str):
            method_name = name
        else:
            # Trying to facilitate tmpl_label_1 that is used in FormNodes,
            # where 1 refers to self.bound_fields[1].
            # In that case the value of name will be ['label_1', 'label']
            name, method_name = name
        slot_name = f'tmpl_{name}'
        if slot_name in slots:
            return

        method = getattr(self, f'render_tmpl_{method_name}')
        slots[slot_name] = method(values, context)


    def format(self, tpl, values, context=None):
        """Apply the prepared values to the templates.
        """
        # pylint:disable=too-many-nested-blocks
        if context:
            # Assume the caller want to tell us there are sub-templates.
            slots = {}
            for typ, nam in SLOT_NAME_PATTERN.findall(tpl):
                tplvalues = values
                if typ == 'tmpl':
                    # Trying to facilitate tmpl_label_1 that is used in
                    # FormNodes, where 1 refers to self.bound_fields[1].
                    match = TMPL_MULTI_PATTERN.match(nam)
                    if match:
                        index = int(match.group('index'))
                        name = match.group('name')
                        for ii, field in enumerate(self.bound_fields):
                            if ii != index:
                                continue
                            nam = [nam, name]
                            tplvalues = values.copy()
                            tplvalues['_bound_field'] = field
                            break
                method = getattr(self, typ)
                method(nam, tplvalues, context, slots)
            if slots:
                tpl = tpl.format_map(IgnoreMissing(slots))

        return tpl.format_map(IgnoreMissing(values))


    def render_tmpl_element(self, values, context):
        """Render django form fields.
        """
        bound_field = values['_bound_field']

        attrs = dict(values['props_raw'])
        attrs['class'] = bound_field.field.widget.attrs.get('class', '').split()
        if bound_field.help_text or 'help' in self.slots:
            attrs['aria-controls'] = 'hint-' + values['id']
            attrs['aria-describedby'] = 'hint-' + values['id']

        self.prepare_element_props(attrs, context, bound_field)
        attrs['class'] = ' '.join(attrs['class'])

        if var_eval(self.kwargs.get('hidden'), context):
            return bound_field.as_hidden(attrs=attrs)

        widget = self.eval(self.kwargs.get('widget'), context)
        if widget == 'input':
            widget = CustomTextInput(attrs=attrs)
            return bound_field.as_widget(widget=widget, attrs=attrs)
        return bound_field.as_widget(attrs=attrs)


    def render_tmpl_errors(self, values, context):
        """Dynamically render a part of the component's template.
        """
        tmpl_t = '<div class="bx--form-requirement__title">{title}.</div>'
        tmpl_s = '<p class="bx--form-requirement__supplement">{supplement}</p>'

        items = []
        bound_field = values['_bound_field']
        for error in bound_field.errors:
            title, supplement = error.split('.', 1)
            supplement = supplement.strip()
            items.append(tmpl_t.format(title=title))
            if supplement:
                items.append(tmpl_s.format(supplement=supplement.strip()))
        return '\n'.join(items)


class ChoiceFormNode(FormNode):
    """Base class for a single choice from form field with multiple choices.

    The first argument to the tag is the Django form field.
    """
    BASE_NODE_PROPS = ('value', *FormNode.BASE_NODE_PROPS)
    "Base Template Tag arguments."

    value = None
    widget_class = CustomCheckboxInput

    def default_id(self):
        """Get Django form field html id.
        """
        # Somehow this ended up here.
        # If the user didn't set the value argument, we assume they wanted us to
        # take it from the bound field.
        # There is strong assumption that this method is called before the label
        # method below.
        if self.value is None:
            if isinstance(self.bound_value, str):
                self.value = self.bound_value
            elif isinstance(self.bound_value, Sequence):
                self.value = self.bound_value[0]
            else:
                self.value = ''

        id_ = self.bound_field.id_for_label
        for ii, (_, val, _) in enumerate(self.choices()):
            if val == self.value:
                return f'{id_}-{ii}'
        return id_


    def label(self):
        """Get Django form field label.
        """
        for _, val, txt in self.choices():
            if val == self.value:
                return txt
        return ''


    def before_prepare(self, values, context):
        """Initialize the values meant for rendering templates.
        """
        self.value = self.eval(self.kwargs.get('value', ''), context)
        super().before_prepare(values, context)

        self.bound_field.field.widget = self.widget_class(
                self.bound_field.field.widget.attrs,
                check_test=self.check_test)

        values['props'].append(('id', self._id))
        values['props'].append(('value', self.value))


    def check_test(self, value):
        """ Test checked attribute

        check_test is a callable that takes a value and returns True if the
        checkbox should be checked for that value.
        """
        if not value:
            return False
        if isinstance(self.bound_value, str):
            return value == self.bound_value
        if isinstance(self.bound_value, Sequence):
            return value in self.bound_value
        return False


class DumbFormNode(Node):
    """For rendering html elements without the associated Django form fields.

    Experimental.
    """
    SLOTS = ('help', 'icon')
    "Named children."
    BASE_NODE_PROPS = ('hidden', 'disabled', *Node.BASE_NODE_PROPS)
    "Base Template Tag arguments."

    def element(self, values, context):
        """Render html of the form control.
        """
        attrs = dict(values['props_raw'])
        attrs['class'] = []
        if 'help' in self.slots:
            attrs['aria-controls'] = 'hint-' + values['id']
            attrs['aria-describedby'] = 'hint-' + values['id']

        self.prepare_element_props(attrs, attrs, context)
        attrs['class'] = ' '.join(attrs['class'])

        if var_eval(self.kwargs.get('hidden'), context):
            attrs.pop('type')
            props = self.join_attributes(self.prune_attributes(attrs))
            return '<input type="hidden" {props}">'.format(props=props)
        return self.render_form_control(attrs, context)


    def render_form_control(self, values, context):
        """Default to rendering input, need to be overridden by subclass.
        """
        props = self.join_attributes(self.prune_attributes(values))
        return '<input {props}">'.format(props=props)


    def after_prepare(self, values, context):
        """Post-process the values for rendering templates.
        """
        values['element'] = self.element(values, context)
        super().after_prepare(values, context)


    def prepare_element_props(self, props, default, context):
        """Prepare html attributes for rendering the form element.
        """


    def render_slot_help(self, values, context):
        """Render html of the slot.
        """
        tmpl = """
<div id="hint-{id}" class="bx--form__helper-text {class}" {props}>
  {child}
</div>
"""
        return tmpl.format(**values)


    def render_slot_icon(self, values, context):
        """Render html of the slot.
        """
        tmpl = """
<svg focusable="false" preserveAspectRatio="xMidYMid meet"
    style="will-change: transform;" xmlns="http://www.w3.org/2000/svg"
    class="bx--btn__icon {class}" width="16" height="16" viewBox="0 0 16 16"
    aria-hidden="true" {props}>
  {child}
</svg>
"""
        return tmpl.format(**values)


components = {
    'Slot': Slot,
}



"""
Tooltip
=======

See: https://www.carbondesignsystem.com/components/tooltip/usage/

Tooltips display additional information upon click, hover, or focus. The
information should be contextual, useful, and nonessential.

Overview
--------

A tooltip is a message box that is displayed when a user hovers over, clicks
or gives focus to a UI element such as an icon, a highlighted word, or a button.
""" # pylint:disable=line-too-long
# pylint:disable=too-many-lines

from .base import Node, modify_svg

class BaseTooltip(Node):
    """Base tooltip.
    """
    WANT_CHILDREN = True
    "Template Tag needs closing end tag."
    NODE_PROPS = ('id', 'align', 'position')
    "Extended Template Tag arguments."
    POSSIBLE_ALIGN = ('start', 'center', 'end')
    "Documentation only."
    POSSIBLE_POSITION = ('top', 'right', 'bottom', 'left')
    "Documentation only."

    def prepare(self, values, context):
        """Prepare values for rendering the templates.
        """
        align = self.eval(self.kwargs.get('align'), context)
        if align and align in self.POSSIBLE_ALIGN:
            values['class'].append(f'bx--tooltip--align-{align}')

        position = self.eval(self.kwargs.get('position'), context)
        if position and position in self.POSSIBLE_POSITION:
            values['class'].append(f'bx--tooltip--{position}')


    def render_slot_icon(self, values, context):
        """Render html of the slot.
        """
        return modify_svg(values['child'], {
            'focusable': 'false',
            'preserveAspectRatio': 'xMidYMid meet',
            'fill': 'currentColor',
            'style': {
                'width': '%spx' % 16,
                'height': '%spx' % 16,
            },
            'aria-hidden': 'true',
        })


class Interactive(BaseTooltip):
    """Tooltip component.
    """
    SLOTS = ('icon', 'footer', 'heading')
    "Named children."
    MODES = ('interactive', 'nolabel')
    "Available variants."
    REQUIRED_PROPS = ('label',)
    "Will raise Exception if not set."
    CLASS_AND_PROPS = ('label', 'content')
    "Prepare xxx_class and xxx_props values."

    def prepare(self, values, context):
        """Prepare values for rendering the templates.
        """
        if 'heading' in self.slots:
            values['content_props'].append(
                    ('aria-labelledby', 'heading-' + self._id))
        elif self.mode == 'nolabel':
            values['content_props'].append(('aria-label', values['label']))
        else:
            values['content_props'].append(
                    ('aria-labelledby', 'label-' + self._id))


    def render_interactive(self, values, context):
        """Output html of the component.
        """
        template = """
<div id="label-{id}" class="bx--tooltip__label">
  {label}
  <button aria-expanded="false" aria-labelledby="label-{id}"
      data-tooltip-trigger data-tooltip-target="#{id}"
      class="bx--tooltip__trigger {class}" aria-controls="{id}" {props}>
    {slot_icon}
  </button>
</div>
<div id="{id}" aria-hidden="true" data-floating-menu-direction="bottom"
    class="bx--tooltip">
  <span class="bx--tooltip__caret"></span>
  <div class="bx--tooltip__content {content_class}" tabindex="-1" role="dialog"
      aria-describedby="body-{id}" {content_props}>
    {slot_heading}
    <p id="body-{id}">{child}</p>
    {slot_footer}
  </div>
  <span tabindex="0"></span>
</div>
"""
        return self.format(template, values, context)


    def render_nolabel(self, values, context):
        """Output html of the component.
        """
        template = """
<div id="label-{id}" class="bx--tooltip__label">
  {label}
  <div tabindex="0" aria-expanded="false" aria-labelledby="label-{id}"
      data-tooltip-trigger data-tooltip-target="#{id}"
      role="button" class="bx--tooltip__trigger {class}" aria-controls="{id}" {props}>
    {slot_icon}
  </div>
</div>
<div id="{id}" aria-hidden="true" data-floating-menu-direction="bottom"
    class="bx--tooltip">
  <span class="bx--tooltip__caret"></span>
  <div class="bx--tooltip__content {content_class}" tabindex="-1" role="dialog"
      aria-describedby="body-{id}" {content_props}>
    {slot_heading}
    <p id="body-{id}">{child}</p>
    {slot_footer}
  </div>
  <span tabindex="0"></span>
</div>
"""
        return self.format(template, values, context)


    def render_slot_heading(self, values, context):
        """Render html of the slot.
        """
        template = """
<h4 id="heading-{id}" class="bx--tooltip__heading {class}" {props}>{child}</h4>
"""
        return self.format(template, values)


    def render_slot_footer(self, values, context):
        """Render html of the slot.
        """
        template = """
<div class="bx--tooltip__footer {class}" {props}>
  {child}
</div>
"""
        return self.format(template, values)


class Definition(BaseTooltip):
    """Tooltip component.
    """
    REQUIRED_PROPS = ('label',)
    "Will raise Exception if not set."

    def render_default(self, values, context):
        """Output html of the component.
        """
        template = """
<div class="bx--tooltip--definition bx--tooltip--a11y" data-tooltip-definition>
  <button aria-describedby="{id}"
      class="bx--tooltip__trigger bx--tooltip--a11y bx--tooltip__trigger--definition {class}"
      {props}>
    {label}
  </button>
  <div class="bx--assistive-text" id="{id}" role="tooltip">
    {child}
  </div>
</div>
"""
        return self.format(template, values)


class Icon(BaseTooltip):
    """Tooltip component.
    """
    SLOTS = ('icon',)
    "Named children."

    def render_default(self, values, context):
        """Output html of the component.
        """
        template = """
<button class="bx--tooltip__trigger bx--tooltip--a11y {class}"
    data-tooltip-icon {props}>
  <span class="bx--assistive-text">{child}</span>
  {slot_icon}
</button>
"""
        return self.format(template, values, context)


components = {
    'InteractiveTooltip': Interactive,
    'DefinitionTooltip': Definition,
    'IconTooltip': Icon,
}



"""
Number Input
============

See: https://www.carbondesignsystem.com/components/number-input/usage/

The number input component is used for entering numeric values and includes
controls for incrementally increasing or decreasing the value.

Overview
--------

Number inputs are similar to text inputs, but contain controls used to
increase or decrease an incremental value.
""" # pylint:disable=line-too-long
# pylint:disable=too-many-lines

from django.utils.translation import gettext as _
#-
from .base import FormNode

class NumberInput(FormNode):
    """Number Input component.
    """
    MODES = ('default', 'mobile')
    "Available variants."
    NODE_PROPS = ('nolabel', 'light')
    "Extended Template Tag arguments."

    def prepare(self, values, context):
        """Prepare values for rendering the templates.
        """
        values['txt_increase'] = _("increase number input")
        values['txt_decrease'] = _("decrease number input")

        if self.eval(self.kwargs.get('nolabel'), context):
            values['class'].append('bx--number--nolabel')

        if self.eval(self.kwargs.get('light'), context):
            values['class'].append('bx--number--light')

        if self.bound_field.help_text:
            values['class'].append('bx--number--helpertext')


    def prepare_element_props(self, props, context):
        """Prepare html attributes for rendering the form element.
        """
        if self.bound_field.errors:
            props['role'] = 'alert'
            props['aria-atomic'] = 'true'


    def render_default(self, values, context):
        """Output html of the component.
        """
        if self.bound_field.errors:
            template = """
<div class="bx--form-item">
  <div data-invalid data-numberinput class="bx--number {class}">
    {tmpl_label}
    <div class="bx--number__input-wrapper">
      {tmpl_element}
      {tmpl_icon_invalid}
      <div class="bx--number__controls">
        {tmpl_btn_incr}
        {tmpl_btn_decr}
      </div>
    </div>
    <div class="bx--form-requirement">
      {tmpl_errors}
    </div>
    {tmpl_help}
  </div>
</div>
"""
        else:
            template = """
<div class="bx--form-item">
  <div data-numberinput class="bx--number {class}">
    {tmpl_label}
    <div class="bx--number__input-wrapper">
      {tmpl_element}
      <div class="bx--number__controls">
        {tmpl_btn_incr}
        {tmpl_btn_decr}
      </div>
    </div>
    {tmpl_help}
  </div>
</div>
"""
        return self.format(template, values, context)


    def render_mobile(self, values, context):
        """Output html of the component.
        """
        if self.bound_field.errors:
            template = """
<div class="bx--form-item">
  <div data-invalid data-numberinput
      class="bx--number bx--number--mobile {class}">
    {tmpl_label}
    <div class="bx--number__input-wrapper">
      {tmpl_btn_decr}
      {tmpl_element}
      {tmpl_btn_incr}
    </div>
    <div class="bx--form-requirement">
      {tmpl_errors}
    </div>
    {tmpl_help}
  </div>
</div>
"""
        else:
            template = """
<div class="bx--form-item">
  <div data-numberinput class="bx--number bx--number--mobile {class}">
    {tmpl_label}
    <div class="bx--number__input-wrapper">
      {tmpl_btn_decr}
      {tmpl_element}
      {tmpl_btn_incr}
    </div>
    {tmpl_help}
  </div>
</div>
"""
        return self.format(template, values, context)


    def render_tmpl_label(self, values, context):
        """Dynamically render a part of the component's template.
        """
        if self.eval(self.kwargs.get('nolabel'), context):
            return ''

        template = """
<label for="{id}" class="bx--label {label_class}" {label_props}>
  {label}
</label>
"""
        return self.format(template, values)


    def render_tmpl_icon_invalid(self, values, context):
        """Dynamically render a part of the component's template.
        """
        return """
<svg focusable="false" preserveAspectRatio="xMidYMid meet"
    xmlns="http://www.w3.org/2000/svg" fill="currentColor"
    class="bx--number__invalid" width="16" height="16" viewBox="0 0 16 16"
    aria-hidden="true">
  <path d="M8,1C4.2,1,1,4.2,1,8s3.2,7,7,7s7-3.1,7-7S11.9,1,8,1z M7.5,4h1v5h-1C7.5,9,7.5,4,7.5,4z M8,12.2    c-0.4,0-0.8-0.4-0.8-0.8s0.3-0.8,0.8-0.8c0.4,0,0.8,0.4,0.8,0.8S8.4,12.2,8,12.2z"></path>
  <path d="M7.5,4h1v5h-1C7.5,9,7.5,4,7.5,4z M8,12.2c-0.4,0-0.8-0.4-0.8-0.8s0.3-0.8,0.8-0.8 c0.4,0,0.8,0.4,0.8,0.8S8.4,12.2,8,12.2z" data-icon-path="inner-path" opacity="0"></path>
</svg>
"""


    def render_tmpl_btn_incr(self, values, context):
        """Dynamically render a part of the component's template.
        """
        template = """
<button aria-label="{txt_increase}" class="bx--number__control-btn up-icon"
    type="button" aria-live="polite" aria-atomic="true">
  <svg focusable="false" preserveAspectRatio="xMidYMid meet"
      xmlns="http://www.w3.org/2000/svg" fill="currentColor" width="8"
      height="4" viewBox="0 0 8 4" aria-hidden="true">
    <path d="M0 4L4 0 8 4z"></path>
  </svg>
</button>
"""
        return self.format(template, values)


    def render_tmpl_btn_decr(self, values, context):
        """Dynamically render a part of the component's template.
        """
        template = """
<button aria-label="{txt_decrease}" class="bx--number__control-btn down-icon"
    type="button" aria-live="polite" aria-atomic="true">
  <svg focusable="false" preserveAspectRatio="xMidYMid meet"
      xmlns="http://www.w3.org/2000/svg" fill="currentColor" width="8"
      height="4" viewBox="0 0 8 4" aria-hidden="true">
    <path d="M8 0L4 4 0 0z"></path>
  </svg>
</button>
"""
        return self.format(template, values)


components = {
    'NumberInput': NumberInput,
}



"""
Search
======

See: https://www.carbondesignsystem.com/components/search/usage/

Search enables users to specify a word or a phrase to find relevant pieces of
content without the use of navigation.

Overview
--------

Search offers users a way to explore a website or application using keywords.
Search can be used as the primary means of discovering content, or as a filter
to aid the user in finding content.
""" # pylint:disable=line-too-long
# pylint:disable=too-many-lines

from django.utils.translation import gettext as _
#-
from .base import Node

class Search(Node):
    """Search component.
    """
    NODE_PROPS = ('id', 'size', 'light')
    "Extended Template Tag arguments."
    CLASS_AND_PROPS = ('label',)
    "Prepare xxx_class and xxx_props values."
    POSSIBLE_SIZE = ('sm', 'lg', 'xl')
    "Documentation only."

    size = None

    def prepare(self, values, context):
        """Prepare values for rendering the templates.
        """
        values['txt_clear'] = _("Clear search input")

        if not values['label']:
            values['label'] = _("Search")

        self.size = size = self.eval(self.kwargs.get('size', 'sm'), context)
        if self.size:
            values['class'].append(f'bx--search--{size}')

        if self.size == 'xl':
            values['close_icon_width'] = 20
        else:
            values['close_icon_width'] = 16

        if self.eval(self.kwargs.get('light'), context):
            values['class'].append('bx--search--light')


    def render_default(self, values, context):
        """Output html of the component.
        """
        template = """
  <div data-search role="search" class="bx--search {class}">
    <label for="{id}" class="bx--label {label_class}" {label_props}>
      {label}
    </label>
    <input class="bx--search-input" type="text" id="{id}"
        placeholder="{label}">
    <svg focusable="false" preserveAspectRatio="xMidYMid meet"
        xmlns="http://www.w3.org/2000/svg" fill="currentColor"
        class="bx--search-magnifier" width="16" height="16" viewBox="0 0 16 16"
        aria-hidden="true">
      <path d="M15,14.3L10.7,10c1.9-2.3,1.6-5.8-0.7-7.7S4.2,0.7,2.3,3S0.7,8.8,3,10.7c2,1.7,5,1.7,7,0l4.3,4.3L15,14.3z M2,6.5 C2,4,4,2,6.5,2S11,4,11,6.5S9,11,6.5,11S2,9,2,6.5z"></path>
    </svg>
    <button class="bx--search-close bx--search-close--hidden"
        title="{txt_clear}" aria-label="{txt_clear}">
      <svg focusable="false" preserveAspectRatio="xMidYMid meet"
          xmlns="http://www.w3.org/2000/svg" fill="currentColor"
          class="bx--search-clear" width="{close_icon_width}"
          height="{close_icon_width}" viewBox="0 0 32 32" aria-hidden="true">
        <path d="M24 9.4L22.6 8 16 14.6 9.4 8 8 9.4 14.6 16 8 22.6 9.4 24 16 17.4 22.6 24 24 22.6 17.4 16 24 9.4z"></path>
      </svg>
    </button>
  </div>
"""
        sm_template = f'<div class="bx--form-item">\n{template}\n</div>'
        if self.size == 'sm':
            return self.format(sm_template, values)
        return self.format(template, values)


components = {
    'Search': Search,
}



"""
Combo Box
=========

See: https://the-carbon-components.netlify.app/?nav=combo-box

""" # pylint:disable=line-too-long
# pylint:disable=too-many-lines

from django.utils.translation import gettext as _
#-
from .base import Node, FormNode

class ComboBox(FormNode):
    """Combo Box component.
    """
    WANT_CHILDREN = True
    "Template Tag needs closing end tag."
    NODE_PROPS = ('light', 'expanded')
    "Extended Template Tag arguments."
    CLASS_AND_PROPS = ('label', 'help', 'combo', 'list', 'icon')
    "Prepare xxx_class and xxx_props values."

    expanded = None

    def prepare(self, values, context):
        """Prepare values for rendering the templates.
        """
        values['txt_open'] = _("Open menu")
        values['txt_close'] = _("Close menu")
        values['txt_clear'] = _("Clear all")

        if self.eval(self.kwargs.get('disabled'), context):
            values['label_class'].append('bx--label--disabled')
            values['help_class'].append('bx--form__helper-text--disabled')
            values['combo_class'].append('bx--list-box--disabled')
            values['list_props'].append(('aria-haspopup', 'true'))
            values['list_props'].append(('disabled', True))
        else:
            values['list_props'].append(('aria-haspopup', 'listbox'))

        if self.eval(self.kwargs.get('light'), context):
            values['combo_class'].append('bx--list-box--light')

        self.expanded = self.eval(self.kwargs.get('expanded'), context)
        if self.expanded:
            values['txt_menu'] = values['txt_close']
            values['combo_class'].append('bx--list-box--expanded')
            values['list_props'].append(('aria-expanded', 'true'))
            values['icon_class'].append('bx--list-box__menu-icon--open')
        else:
            values['txt_menu'] = values['txt_open']
            values['list_props'].append(('aria-expanded', 'false'))


    def prepare_element_props(self, props, context):
        """Prepare html attributes for rendering the form element.
        """
        props['class'].append('bx--text-input')
        props['aria-autocomplete'] = 'list'
        props['autocomplete'] = 'off'
        props['aria-owns'] = 'menu-' + self._id

        if self.expanded:
            props['aria-expanded'] = 'true'
        else:
            props['aria-expanded'] = 'false'


    def render_default(self, values, context):
        """Output html of the component.
        """
        if self.bound_field.errors:
            template = """
<div class="bx--form-item">
  <div class="bx--list-box__wrapper">
    {tmpl_label}
    <div class="bx--combo-box bx--list-box {combo_class}" data-invalid>
      <div role="combobox" class="bx--list-box__field" aria-label="{txt_menu}"
          {list_props}>
        {tmpl_element}
        {tmpl_icon_invalid}
        {tmpl_icon_clear}
        {tmpl_icon_menu}
      </div>
      <ul class="bx--list-box__menu" role="listbox" id="menu-{id}"
          aria-label="{label}">
        {child}
      </ul>
    </div>
    {tmpl_help}
    <div class="bx--form-requirement">
      {tmpl_errors}
    </div>
  </div>
</div>
"""
        else:
            template = """
<div class="bx--form-item">
  <div class="bx--list-box__wrapper">
    {tmpl_label}
    <div class="bx--combo-box bx--list-box {combo_class}">
      <div role="combobox" class="bx--list-box__field" aria-label="{txt_menu}"
          {list_props}>
        {tmpl_element}
        {tmpl_icon_clear}
        {tmpl_icon_menu}
      </div>
      <ul class="bx--list-box__menu" role="listbox" id="menu-{id}"
          aria-label="{label}">
        {child}
      </ul>
    </div>
    {tmpl_help}
  </div>
</div>
"""
        return self.format(template, values, context)


    def render_tmpl_icon_clear(self, values, context):
        """Dynamically render a part of the component's template.
        """
        if not self.bound_value:
            return ''
        return """
<div class="bx--list-box__selection" role="button" title="{txt_clear}">
  <svg focusable="false" preserveAspectRatio="xMidYMid meet"
      xmlns="http://www.w3.org/2000/svg" fill="currentColor"
      aria-label="{txt_clear}" width="16" height="16" viewBox="0 0 32 32"
      role="img">
    <path d="M24 9.4L22.6 8 16 14.6 9.4 8 8 9.4 14.6 16 8 22.6 9.4 24 16 17.4 22.6 24 24 22.6 17.4 16 24 9.4z"></path>
  </svg>
</div>
"""


    def render_tmpl_icon_menu(self, values, context):
        """Dynamically render a part of the component's template.
        """
        return """
<div class="bx--list-box__menu-icon {icon_class}" {icon_props}>
  <svg focusable="false" preserveAspectRatio="xMidYMid meet"
      xmlns="http://www.w3.org/2000/svg" fill="currentColor"
      aria-label="{txt_menu}" width="16" height="16" viewBox="0 0 16 16"
      role="img">
    <path d="M8 11L3 6 3.7 5.3 8 9.6 12.3 5.3 13 6z"></path>
  </svg>
</div>
"""


    def render_tmpl_icon_invalid(self, values, context):
        """Dynamically render a part of the component's template.
        """
        return """
<svg focusable="false" preserveAspectRatio="xMidYMid meet"
    xmlns="http://www.w3.org/2000/svg" fill="currentColor"
    class="bx--list-box__invalid-icon" width="16" height="16"
    viewBox="0 0 16 16" aria-hidden="true">
  <path d="M8,1C4.2,1,1,4.2,1,8s3.2,7,7,7s7-3.1,7-7S11.9,1,8,1z M7.5,4h1v5h-1C7.5,9,7.5,4,7.5,4z M8,12.2   c-0.4,0-0.8-0.4-0.8-0.8s0.3-0.8,0.8-0.8c0.4,0,0.8,0.4,0.8,0.8S8.4,12.2,8,12.2z"></path>
  <path d="M7.5,4h1v5h-1C7.5,9,7.5,4,7.5,4z M8,12.2c-0.4,0-0.8-0.4-0.8-0.8s0.3-0.8,0.8-0.8 c0.4,0,0.8,0.4,0.8,0.8S8.4,12.2,8,12.2z" data-icon-path="inner-path" opacity="0"></path>
</svg>
"""


class ComboBoxItem(Node):
    """Combo Box item.
    """
    WANT_CHILDREN = True
    "Template Tag needs closing end tag."
    NODE_PROPS = ('active',)
    "Extended Template Tag arguments."

    def prepare(self, values, context):
        """Prepare values for rendering the templates.
        """
        if self.eval(self.kwargs.get('active'), context):
            values['class'].append('bx--list-box__menu-item--highlighted')


    def render_default(self, values, context):
        """Output html of the component.
        """
        template = """
<li class="bx--list-box__menu-item {class}" {props}>
  <div class="bx--list-box__menu-item__option" tabindex="0">
    {child}
  </div>
</li>
"""
        return self.format(template, values)


components = {
    'ComboBox': ComboBox,
    'ComboBoxItem': ComboBoxItem,
}



"""
File Uploader
=============

See: https://www.carbondesignsystem.com/components/file-uploader/usage/

File uploaders allow users to select one or more files to upload to a specific
location.

Overview
--------

File uploaders allow users to upload content of their own. A file uploader is
commonly found in forms, but can also live as a standalone element. There are
two variants of file uploaders—our default file uploader and a drag and drop
file uploader.
""" # pylint:disable=line-too-long
# pylint:disable=too-many-lines

from django.utils.translation import gettext as _
#-
from .base import FormNode

class FileUploader(FormNode):
    """File Uploader component.
    """
    NODE_PROPS = ('is_loading',)
    "Extended Template Tag arguments."

    def prepare(self, values, context):
        """Prepare values for rendering the templates.
        """
        values['txt_drop'] = _("Drag and drop files here or upload")
        values['txt_clear'] = _("Remove uploaded file")


    def prepare_element_props(self, props, context):
        """Prepare html attributes for rendering the form element.
        """
        props['class'].append('bx--file-input')
        props['data-file-uploader'] = ''
        props['data-target'] = '#container-' + self._id


    def render_default(self, values, context):
        """Output html of the component.
        """
        values['filename'] = self.bound_field.value() or ''

        if self.bound_field.errors:
            template = """
<div class="bx--form-item">
  <strong class="bx--file--label {label_class}" {label_props}>
    {label}
  </strong>
  {tmpl_help}
  <div class="bx--file" data-file>
    <label for="{id}" class="bx--file-browse-btn" role="button" tabindex="0">
      <div data-file-drop-container class="bx--file__drop-container">
        {txt_drop}
        {tmpl_element}
      </div>
    </label>
    <div data-file-container id="container-{id}" class="bx--file-container">
      <div class="bx--file__selected-file bx--file__selected-file--invalid"
          data-invalid>
        <p class="bx--file-filename">{filename}</p>
        <span data-for="prepopulated-file-uploader"
            class="bx--file__state-container">
          <svg focusable="false" preserveAspectRatio="xMidYMid meet"
              xmlns="http://www.w3.org/2000/svg" fill="currentColor"
              class="bx--file--invalid" width="16" height="16"
              viewBox="0 0 16 16" aria-hidden="true">
            <path d="M8,1C4.2,1,1,4.2,1,8s3.2,7,7,7s7-3.1,7-7S11.9,1,8,1z M7.5,4h1v5h-1C7.5,9,7.5,4,7.5,4z M8,12.2  c-0.4,0-0.8-0.4-0.8-0.8s0.3-0.8,0.8-0.8c0.4,0,0.8,0.4,0.8,0.8S8.4,12.2,8,12.2z"></path>
            <path d="M7.5,4h1v5h-1C7.5,9,7.5,4,7.5,4z M8,12.2c-0.4,0-0.8-0.4-0.8-0.8s0.3-0.8,0.8-0.8 c0.4,0,0.8,0.4,0.8,0.8S8.4,12.2,8,12.2z" data-icon-path="inner-path" opacity="0"></path>
          </svg>
          <svg focusable="true" preserveAspectRatio="xMidYMid meet"
              xmlns="http://www.w3.org/2000/svg" fill="currentColor"
              aria-label="{txt_clear}" class="bx--file-close"
              width="16" height="16" viewBox="0 0 32 32" role="img"
              tabindex="0">
            <path d="M24 9.4L22.6 8 16 14.6 9.4 8 8 9.4 14.6 16 8 22.6 9.4 24 16 17.4 22.6 24 24 22.6 17.4 16 24 9.4z"></path>
          </svg>
        </span>
        <div class="bx--form-requirement">
          {tmpl_errors}
        </div>
      </div>
    </div>
  </div>
</div>
"""
        elif self.eval(self.kwargs.get('is_loading'), context):
            template = """
<div class="bx--form-item">
  <strong class="bx--file--label {label_class}" {label_props}>
    {label}
  </strong>
  {tmpl_help}
  <div class="bx--file" data-file>
    <label for="{id}" class="bx--file-browse-btn" role="button" tabindex="0">
      <div data-file-drop-container class="bx--file__drop-container">
        {txt_drop}
        {tmpl_element}
      </div>
    </label>
    <div data-file-container id="container-{id}" class="bx--file-container">
      <div class="bx--file__selected-file">
        <p class="bx--file-filename">{filename}</p>
        <span data-for="prepopulated-file-uploader" class="bx--file__state-container">
          <div class="bx--inline-loading__animation">
            <div data-inline-loading-spinner="" class="bx--loading bx--loading--small">
              <svg class="bx--loading__svg" viewBox="-75 -75 150 150">
                <circle class="bx--loading__background" cx="0" cy="0" r="26.8125"></circle>
                <circle class="bx--loading__stroke" cx="0" cy="0" r="26.8125"></circle>
              </svg>
            </div>
          </div>
        </span>
      </div>
    </div>
  </div>
</div>
"""
        elif values['filename']:
            template = """
<div class="bx--form-item">
  <strong class="bx--file--label {label_class}" {label_props}>
    {label}
  </strong>
  {tmpl_help}
  <div class="bx--file" data-file>
    <label for="{id}" class="bx--file-browse-btn" role="button" tabindex="0">
      <div data-file-drop-container class="bx--file__drop-container">
        {txt_drop}
        {tmpl_element}
      </div>
    </label>
    <div data-file-container id="container-{id}" class="bx--file-container">
      <div class="bx--file__selected-file">
        <p class="bx--file-filename">{filename}</p>
        <span data-for="prepopulated-file-uploader"
            class="bx--file__state-container">
          <svg focusable="false" preserveAspectRatio="xMidYMid meet"
              xmlns="http://www.w3.org/2000/svg" fill="currentColor"
              class="bx--file-complete" width="16" height="16"
              viewBox="0 0 16 16" aria-hidden="true">
            <path d="M8,1C4.1,1,1,4.1,1,8c0,3.9,3.1,7,7,7s7-3.1,7-7C15,4.1,11.9,1,8,1z M7,11L4.3,8.3l0.9-0.8L7,9.3l4-3.9l0.9,0.8L7,11z"></path>
            <path d="M7,11L4.3,8.3l0.9-0.8L7,9.3l4-3.9l0.9,0.8L7,11z" data-icon-path="inner-path" opacity="0"></path>
          </svg>
        </span>
      </div>
    </div>
  </div>
</div>
"""
        else:
            template = """
<div class="bx--form-item">
  <strong class="bx--file--label {label_class}" {label_props}>
    {label}
  </strong>
  {tmpl_help}
  <div class="bx--file" data-file>
    <label for="{id}" class="bx--file-browse-btn" role="button" tabindex="0">
      <div data-file-drop-container class="bx--file__drop-container">
        {txt_drop}
        {tmpl_element}
      </div>
    </label>
    <div data-file-container id="container-{id}" class="bx--file-container">
    </div>
  </div>
</div>
"""
        return self.format(template, values, context)


    def render_tmpl_help(self, values, context):
        """Dynamically render a part of the component's template.
        """
        if self.bound_field.help_text:
            tmpl = """
<p id="hint-{id}" class="bx--label-description {class}" {props}>
  {child}
</p>
"""
            help_values = {
                'child': self.bound_field.help_text,
                'id': values['id'],
                'class': values['help_class'],
                'props': values['help_props'],
            }
            return tmpl.format(**help_values)
        return ''


components = {
    'File': FileUploader,
}



"""
Content Switcher
================

See: https://www.carbondesignsystem.com/components/content-switcher/usage/

Content switchers allow users to toggle between two or more content sections
within the same space on screen.

Overview
--------

Content switchers allow users to toggle between alternate views of similar or
related content. Only one content section is shown at a time.
""" # pylint:disable=line-too-long
# pylint:disable=too-many-lines

from .base import Node

class ContentSwitcher(Node):
    """Content Switcher component.
    """
    WANT_CHILDREN = True
    "Template Tag needs closing end tag."

    def prepare(self, values, context):
        """Prepare values for rendering the templates.
        """
        if values['label']:
            values['props'].append(('aria-label', values['label']))


    def render_default(self, values, context):
        """Output html of the component.
        """
        template = """
<div data-content-switcher class="bx--content-switcher {class}" role="tablist"
    {props}>
  {child}
</div>
        """
        return self.format(template, values)


class ContentSwitcherItem(Node):
    """Content Switcher item.
    """
    WANT_CHILDREN = True
    "Template Tag needs closing end tag."
    NODE_PROPS = ('target', 'active', 'disabled')
    "Extended Template Tag arguments."
    DEFAULT_TAG = 'button'
    "Rendered HTML tag."

    def prepare(self, values, context):
        """Prepare values for rendering the templates.
        """
        values['target'] = self.eval(self.kwargs['target'], context)

        if self.eval(self.kwargs.get('active'), context):
            values['class'].append('bx--content-switcher--selected')
            values['props'].append(('aria-selected', 'true'))

        if self.eval(self.kwargs.get('disabled'), context):
            values['props'].append(('disabled', True))


    def render_default(self, values, context):
        """Output html of the component.
        """
        template = """
<{tag} class="bx--content-switcher-btn {class}" data-target="{target}"
    role="tab" {props}>
  <span class="bx--content-switcher__label">{child}</span>
</{tag}>
"""
        return self.format(template, values)


components = {
    'ContentSwitcher': ContentSwitcher,
    'ContentSwitcherItem': ContentSwitcherItem,
}



"""
Pagination
==========

See: https://www.carbondesignsystem.com/components/pagination/usage/

Pagination is used for splitting up content or data into several pages, with
a control for navigating to the next or previous page.

Overview
--------

Generally, pagination is used if there are more than 25 items displayed in
one view.

The default number displayed will vary depending on the context.
""" # pylint:disable=line-too-long
# pylint:disable=too-many-lines

import logging
#-
from django.utils.translation import gettext as _
#-
from .base import Node

_logger = logging.getLogger(__name__)


class Pagination(Node):
    """Pagination component.
    """
    NODE_PROPS = ('pager', 'pager_sizes', 'disabled', 'page_name',
            'pagesize_name')
    "Extended Template Tag arguments."
    CLASS_AND_PROPS = ('navbtn',)
    "Prepare xxx_class and xxx_props values."
    PAGER_SIZES = (10, 20, 30, 40, 50)

    pager = None
    pagesize = 10

    def prepare(self, values, context):
        """Prepare values for rendering the templates.
        """
        request = context.get('request')
        self.pager = self.eval(self.kwargs.get('pager'), context)

        values['txt_per_page'] = _("Items per page")
        values['txt_select_per_page'] = _("select number of items per page")
        values['txt_select_page_num'] = _("select page number to view")
        values['txt_back_btn'] = _("previous page")
        values['txt_forw_btn'] = _("next page")

        values['page_name'] = self.eval(self.kwargs.get('page_name', 'page'),
                context)
        values['pagesize_name'] = self.eval(self.kwargs.get('pagesize_name',
                'pagesize'), context)

        if request:
            self.pagesize = request.GET.get(values['pagesize_name'],
                    self.pagesize)

        if self.eval(self.kwargs.get('disabled'), context):
            values['navbtn_class'].append('bx--pagination__button--no-index')


    def render_default(self, values, context):
        """Output html of the component.
        """
        if not self.pager:
            return ''
        if not self.pager.has_previous and not self.pager.has_next:
            return ''

        template = """
<div class="bx--pagination" data-pagination data-page-name="{page_name}"
    data-pagesize-name="{pagesize_name}">
  <div class="bx--pagination__left">
    <label id="select-{id}-pagination-count-label" class="bx--pagination__text"
        for="select-{id}-pagination-count">
      {txt_per_page}:
    </label>
    <div class="bx--select bx--select--inline bx--select__item-count">
      <select class="bx--select-input" id="select-{id}-pagination-count"
          aria-label="{txt_select_per_page}" tabindex="0"
          data-items-per-page>
        {tmpl_pagination_sizes}
      </select>
      <svg focusable="false" preserveAspectRatio="xMidYMid meet"
          xmlns="http://www.w3.org/2000/svg" fill="currentColor"
          class="bx--select__arrow" width="16" height="16" viewBox="0 0 16 16"
          aria-hidden="true">
        <path d="M8 11L3 6 3.7 5.3 8 9.6 12.3 5.3 13 6z"></path>
      </svg>
    </div>
    <span class="bx--pagination__text">
      {tmpl_pagination_range}
    </span>
  </div>
  <div class="bx--pagination__right">
    <div class="bx--select bx--select--inline bx--select__page-number">
      <select class="bx--select-input" id="select-{id}-pagination-page"
          aria-label="{txt_select_page_num}" tabindex="0"
          data-page-number-input>
        {tmpl_pagination_numbers}
      </select>
      <svg focusable="false" preserveAspectRatio="xMidYMid meet"
          xmlns="http://www.w3.org/2000/svg" fill="currentColor"
          class="bx--select__arrow" width="16" height="16" viewBox="0 0 16 16"
          aria-hidden="true">
        <path d="M8 11L3 6 3.7 5.3 8 9.6 12.3 5.3 13 6z"></path>
      </svg>
    </div>
    <label id="select-{id}-pagination-page-label" class="bx--pagination__text"
        for="select-{id}-pagination-page">
      {tmpl_pagination_num_pages}
    </label>
    <button class="bx--pagination__button bx--pagination__button--backward {navbtn_class}"
        tabindex="0" data-page-backward aria-label="{txt_back_btn}" {navbtn_props}>
      <svg focusable="false" preserveAspectRatio="xMidYMid meet"
          xmlns="http://www.w3.org/2000/svg" fill="currentColor"
          class="bx--pagination__nav-arrow" width="20" height="20"
          viewBox="0 0 32 32" aria-hidden="true">
        <path d="M20 24L10 16 20 8z"></path>
      </svg>
    </button>
    <button class="bx--pagination__button bx--pagination__button--forward {navbtn_class}"
        tabindex="0" data-page-forward aria-label="{txt_forw_btn}" {navbtn_props}>
      <svg focusable="false" preserveAspectRatio="xMidYMid meet"
          xmlns="http://www.w3.org/2000/svg" fill="currentColor"
          class="bx--pagination__nav-arrow" width="20" height="20"
          viewBox="0 0 32 32" aria-hidden="true">
        <path d="M12 8L22 16 12 24z"></path>
      </svg>
    </button>
  </div>
</div>
"""
        return self.format(template, values, context)


    def render_tmpl_pagination_sizes(self, values, context):
        """Dynamically render a part of the component's template.
        """
        template = '<option class="bx--select-option" {props}>{value}</option>'
        options = []

        pager_sizes = self.eval(self.kwargs.get('pager_sizes',
                self.PAGER_SIZES), context)
        if isinstance(pager_sizes, str):
            pager_sizes = [int(x) for x in pager_sizes.split(',')]
        else:
            # Make sure it's a copy
            pager_sizes = list(pager_sizes)
        if self.pagesize not in pager_sizes:
            pager_sizes.insert(0, self.pagesize)

        for value in pager_sizes:
            if value == self.pagesize:
                options.append(template.format(value=value, props='selected'))
            else:
                options.append(template.format(value=value, props=''))
        return '\n' + '\n'.join(options)


    def render_tmpl_pagination_numbers(self, values, context):
        """Dynamically render a part of the component's template.
        """
        template = '<option class="bx--select-option" {props}>{value}</option>'
        options = []

        for value in (x + 1 for x in range(self.pager.paginator.num_pages)):
            if value != self.pager.number:
                options.append(template.format(value=value, props=''))
            else:
                options.append(template.format(value=value, props='selected'))
        return '\n' + '\n'.join(options)


    def render_tmpl_pagination_range(self, values, context):
        """Dynamically render a part of the component's template.
        """
        template = """
<span data-displayed-item-range>{range_start} – {range_end}</span>
"""
        page_range = template.format(
                range_start=self.pager.start_index(),
                range_end=self.pager.end_index())
        template = '<span data-total-items>{total}</span>'
        page_total = template.format(total=self.pager.paginator.count)

        range_of_items = _("{range} of {total} items")
        return range_of_items.format(range=page_range, total=page_total)


    def render_tmpl_pagination_num_pages(self, values, context):
        """Dynamically render a part of the component's template.
        """
        num_pages = _("of {total} pages")
        return num_pages.format(total=self.pager.paginator.num_pages)


components = {
    'Pagination': Pagination,
}



"""Template tags provided by carbondesign module.

Provides built-in Carbon Design components template tags,
svg for loading svg file,
carbondesign_assets for loading css/js files.
"""
import logging
import os
import re
#-
from django import template
from django.conf import settings
from django.utils.safestring import mark_safe
#-
from ..tags import accordion, base, breadcrumb, button, checkbox, code_snippet
from ..tags import combo_box, content_switcher, copy_button, data_table
from ..tags import date_picker, file_uploader, grid, inline_loading
from ..tags import link, list_, loading, modal, multi_select
from ..tags import notification, number_input, overflow_menu, pagination
from ..tags import pagination_nav, progress_indicator, radio_button, search
from ..tags import select, slider, structured_list, tabs, tag, text_area
from ..tags import text_input, tile, time_picker, toggle, toolbar, tooltip
from ..tags import ui_shell, ui_shell_switcher

_logger = logging.getLogger(__name__)
register = template.Library()


MATERIAL_TAGS = {
    **accordion.components,
    **base.components,
    **breadcrumb.components,
    **button.components,
    **checkbox.components,
    **code_snippet.components,
    **combo_box.components,
    **content_switcher.components,
    **copy_button.components,
    **data_table.components,
    **date_picker.components,
    **file_uploader.components,
    **grid.components,
    **inline_loading.components,
    **link.components,
    **list_.components,
    **loading.components,
    **modal.components,
    **multi_select.components,
    **notification.components,
    **number_input.components,
    **overflow_menu.components,
    **pagination.components,
    **pagination_nav.components,
    **progress_indicator.components,
    **radio_button.components,
    **search.components,
    **select.components,
    **slider.components,
    **structured_list.components,
    **tabs.components,
    **tag.components,
    **text_area.components,
    **text_input.components,
    **tile.components,
    **time_picker.components,
    **toggle.components,
    **toolbar.components,
    **tooltip.components,
    **ui_shell.components,
    **ui_shell_switcher.components,
}


class TagParser:
    """Parse arguments of Django template tags.
    """
    def __init__(self, tags):
        self.tags = tags

    def __call__(self, parser, token):
        params = token.split_contents()
        tagname = params.pop(0)

        args = []
        kwargs = {}
        for param in params:
            if '=' in param:
                key, val = param.split('=', 1)
            else:
                key, val = (None, param)

            if val[0] in ('"', "'"):
                val = val.strip('"\'')
            else:
                val = template.Variable(val)

            if key:
                kwargs[key] = val
            else:
                args.append(val)

        cls = self.tags[tagname]

        if getattr(cls, 'WANT_CHILDREN', False):
            nodelist = parser.parse(('end' + tagname,))
            parser.delete_first_token()
            args.insert(0, nodelist)

        return cls(*args, **kwargs)


@register.simple_tag
def svg(path):
    """This template tag will load svg files located in settings.SVG_DIRS.
    """
    # Security test.
    if not path or not re.match(r'^\w.*\.svg$', path):
        return ''
    for dirname in settings.SVG_DIRS:
        if not os.path.exists(dirname/path):
            continue
        return mark_safe(open(dirname/path, 'r', encoding='utf-8').read())
    return ''


@register.simple_tag
def carbondesign_assets():
    """This template tag will inject native carbondesign css/js files.

    Don't forget to run collectstatic!
    """
    if settings.DEBUG:
        suffix = ''
    else:
        suffix = '.min'
    tmpl = """
<link rel="stylesheet" type="text/css" href="{baseurl}carbondesign/carbon-components{suffix}.css">
<script src="{baseurl}carbondesign/carbon-components{suffix}.js"></script>
""" # pylint:disable=line-too-long
    return mark_safe(tmpl.format(baseurl=settings.STATIC_URL, suffix=suffix))


_parser = TagParser(MATERIAL_TAGS)
for name in MATERIAL_TAGS:
    register.tag(name, _parser)



"""
Pagination Nav
==============

See: https://the-carbon-components.netlify.app/?nav=pagination-nav

""" # pylint:disable=line-too-long
# pylint:disable=too-many-lines

from django.utils.translation import gettext as _
#-
from .base import Node

NUMBERS_TEMPLATE = """
<li class="bx--pagination-nav__list-item">
  <{item_tag} class="bx--pagination-nav__page {item_class}" data-page="{num}"
      data-page-button {item_props}>
    <span class="bx--pagination-nav__accessibility-label">{txt_page}</span>
    {num}
  </{item_tag}>
</li>
"""
PRE_NUMBERS_SELECT_TEMPLATE = """
<li class="bx--pagination-nav__list-item">
  <div class="bx--pagination-nav__select">
    <select class="bx--pagination-nav__page bx--pagination-nav__page--select"
        data-page-select aria-label="{txt_select}">
      <option value="" hidden></option>
"""
POST_NUMBERS_SELECT_TEMPLATE = """
    </select>
    <div class="bx--pagination-nav__select-icon-wrapper">
      <svg focusable="false" preserveAspectRatio="xMidYMid meet"
          xmlns="http://www.w3.org/2000/svg" fill="currentColor"
          class="bx--pagination-nav__select-icon" width="16" height="16"
          viewBox="0 0 32 32" aria-hidden="true">
        <circle cx="8" cy="16" r="2"></circle>
        <circle cx="16" cy="16" r="2"></circle>
        <circle cx="24" cy="16" r="2"></circle>
      </svg>
    </div>
  </div>
</li>
"""


class PaginationNav(Node):
    """Pagination Nav component.
    """
    NODE_PROPS = ('pager', 'max_links', 'extra_links', 'as_link')
    "Extended Template Tag arguments."
    CLASS_AND_PROPS = ('item',)
    "Prepare xxx_class and xxx_props values."

    pager = None
    links_begin = None
    links_end = None
    links_before = []
    links_after = []

    def calculate_links(self, max_links, max_extra):
        """Prepare page links.
        """
        current = self.pager.number
        last = self.pager.paginator.num_pages

        # Try to keep current page in the middle of the links.
        self.links_begin = max(current - int(max_links / 2), 1)
        self.links_end = min(self.links_begin + max_links - 1, last)

        # Don't show extra selections if the number of links is too low.
        if not max_extra or self.links_end - self.links_begin + 1 < 7:
            return

        # Links 1 and 2 are going to be replaced with first in extra links and
        # the select box.
        begin = max(self.links_begin + 1 - max_extra, 1)
        end = min(self.links_begin + 1, current - 2)
        if begin < end:
            self.links_before = list(range(begin, end + 1))
            self.links_begin = end + 1

        begin = max(self.links_end - 1, current + 2)
        end = min(self.links_end - 1 + max_extra, last)
        if begin < end:
            self.links_after = list(range(begin, end + 1))
            self.links_end = begin - 1


    def prepare(self, values, context):
        """Prepare values for rendering the templates.
        """
        values['txt_pagination'] = _("pagination")
        values['txt_prev'] = _("Previous page")
        values['txt_next'] = _("Next page")
        values['txt_page'] = _("page")
        values['txt_select'] = _("select page number")

        self.pager = self.eval(self.kwargs.get('pager'), context)

        max_links = self.eval(self.kwargs.get('max_links', 7), context)
        max_extra = self.eval(self.kwargs.get('extra_links'), context)
        if self.pager:
            self.calculate_links(max_links, max_extra)

        if self.pager:
            if self.pager.has_previous():
                values['prev_class'] = ''
                values['prev_props'] = ''
            else:
                values['prev_class'] = 'bx--pagination-nav__page--disabled'
                values['prev_props'] = 'aria-disabled="true"'

            if self.pager.has_next():
                values['next_class'] = ''
                values['next_props'] = ''
            else:
                values['next_class'] = 'bx--pagination-nav__page--disabled'
                values['next_props'] = 'aria-disabled="true"'

        if self.eval(self.kwargs.get('as_link'), context):
            values['item_tag'] = 'a'
            values['item_props'].append(('href', 'javascript:void(0)'))
        else:
            values['item_tag'] = 'button'


    def render_default(self, values, context):
        """Output html of the component.
        """
        if not self.pager:
            return ''

        template = """
<nav class="bx--pagination-nav" aria-label="{txt_pagination}" data-pagination-nav>
  <ul class="bx--pagination-nav__list">
    {tmpl_prev}
    {tmpl_numbers_before}
    {tmpl_numbers}
    {tmpl_numbers_after}
    {tmpl_next}
  </ul>
</nav>
"""
        return self.format(template, values, context)


    def render_tmpl_prev(self, values, context):
        """Dynamically render a part of the component's template.
        """
        template = """
<li class="bx--pagination-nav__list-item">
  <{item_tag} class="bx--pagination-nav__page bx--pagination-nav__page--direction {prev_class} {item_class}"
      data-page-previous {prev_props} {item_props}>
    <span class="bx--pagination-nav__accessibility-label">{txt_prev}</span>
    <svg focusable="false" preserveAspectRatio="xMidYMid meet"
        xmlns="http://www.w3.org/2000/svg" fill="currentColor"
        class="bx--pagination-nav__icon" width="5" height="8" viewBox="0 0 5 8"
        aria-hidden="true">
      <path d="M5 8L0 4 5 0z"></path>
    </svg>
  </{item_tag}>
</li>
"""
        return self.format(template, values)


    def render_tmpl_next(self, values, context):
        """Dynamically render a part of the component's template.
        """
        template = """
<li class="bx--pagination-nav__list-item">
  <{item_tag} class="bx--pagination-nav__page bx--pagination-nav__page--direction {next_class} {item_class}"
      data-page-next {next_props} {item_props}>
    <span class="bx--pagination-nav__accessibility-label">{txt_next}</span>
    <svg focusable="false" preserveAspectRatio="xMidYMid meet"
        xmlns="http://www.w3.org/2000/svg" fill="currentColor"
        class="bx--pagination-nav__icon" width="5" height="8" viewBox="0 0 5 8"
        aria-hidden="true">
      <path d="M0 0L5 4 0 8z"></path>
    </svg>
  </{item_tag}>
</li>
"""
        return self.format(template, values)


    def render_tmpl_numbers(self, values, context):
        """Dynamically render a part of the component's template.
        """
        if self.links_begin >= self.links_end:
            return ''
        items = []
        for ii in range(self.links_begin, self.links_end + 1):
            # There are `item_tag` and `txt_page` in the template, but those
            # will be replaced when rendering the whole component.
            options = {'num': ii}
            if ii == self.pager.number:
                options['item_class'] = 'bx--pagination-nav__page--active ' +\
                        'bx--pagination-nav__page--disabled ' +\
                        values['item_class']
                options['item_props'] = 'data-page-active="true" ' +\
                        'aria-current="page" aria-disabled="true" ' +\
                        values['item_props']
            items.append(self.format(NUMBERS_TEMPLATE, options))
        return '\n'.join(items)


    def render_tmpl_numbers_before(self, values, context):
        """Dynamically render a part of the component's template.
        """
        if not self.links_before:
            return ''
        first = self.links_before.pop(0)
        items = []

        items.append(self.format(NUMBERS_TEMPLATE, {'num': first}))
        items.append(PRE_NUMBERS_SELECT_TEMPLATE)
        for ii in self.links_before:
            items.append(f'<option data-page="{ii}">{ii}</option>')
        items.append(POST_NUMBERS_SELECT_TEMPLATE)
        return '\n'.join(items)


    def render_tmpl_numbers_after(self, values, context):
        """Dynamically render a part of the component's template.
        """
        if not self.links_after:
            return ''
        last = self.links_after.pop(-1)
        items = []

        items.append(PRE_NUMBERS_SELECT_TEMPLATE)
        for ii in self.links_after:
            items.append(f'<option data-page="{ii}">{ii}</option>')
        items.append(POST_NUMBERS_SELECT_TEMPLATE)
        items.append(self.format(NUMBERS_TEMPLATE, {'num': last}))
        return '\n'.join(items)


components = {
    'PaginationNav': PaginationNav,
}



"""
Time Picker
===========

See: https://the-carbon-components.netlify.app/?nav=time-picker

""" # pylint:disable=line-too-long
# pylint:disable=too-many-lines

from django.utils.translation import gettext as _
import pytz
#-
from .base import FormNode

class TimePicker(FormNode):
    """Time Picker component.
    """
    NODE_PROPS = ('light', 'zones')
    "Extended Template Tag arguments."
    CLASS_AND_PROPS = ('label', 'select')
    "Prepare xxx_class and xxx_props values."

    def prepare(self, values, context):
        """Prepare values for rendering the templates.
        """
        values['txt_select_ampm'] = _("Select AM/PM")
        values['txt_select_zone'] = _("Select time zone")

        if self.eval(self.kwargs.get('disabled'), context):
            values['label_class'].append('bx--label--disabled')
            values['select_props'].append(('disabled', True))

        if self.eval(self.kwargs.get('light'), context):
            values['class'].append('bx--time-picker--light')
            values['select_class'].append('bx--select--light')


    def prepare_element_props(self, props, context):
        """Prepare html attributes for rendering the form element.
        """
        props['class'].append('bx--text-input')
        props['class'].append('bx--time-picker__input-field')
        props['pattern'] = r'(1[012]|[1-9]):[0-5][0-9](\\s)?'
        props['placeholder'] = 'hh:mm'
        props['maxlength'] = "5"

        if self.eval(self.kwargs.get('light'), context):
            props['class'].append('bx--text-input--light')


    def render_default(self, values, context):
        """Output html of the component.
        """
        light = self.eval(self.kwargs.get('light'), context)
        if self.bound_field.errors:
            if light:
                template = """
<div class="bx--form-item">
  <div class="bx--time-picker {class}" data-invalid>
    <div class="bx--time-picker__input">
      {tmpl_label}
      {tmpl_element}
    </div>
    {tmpl_select_ampm}
    {tmpl_select_zone}
  </div>
  <div class="bx--form-requirement">
    {tmpl_errors}
  </div>
  {tmpl_help}
</div>
"""
            else:
                template = """
<div class="bx--form-item">
  {tmpl_label}
  <div class="bx--time-picker {class}" data-invalid>
    <div class="bx--time-picker__input">{tmpl_element}</div>
    {tmpl_select_ampm}
    {tmpl_select_zone}
  </div>
  <div class="bx--form-requirement">
    {tmpl_errors}
  </div>
  {tmpl_help}
</div>
"""
        elif light and self.eval(self.kwargs.get('disabled'), context):
            template = """
<div class="bx--form-item">
  <div class="bx--time-picker {class}">
    <div class="bx--time-picker__input">
      {tmpl_label}
      {tmpl_element}
    </div>
    {tmpl_select_ampm}
    {tmpl_select_zone}
  </div>
  {tmpl_help}
</div>
"""
        else:
            template = """
<div class="bx--form-item">
  {tmpl_label}
  <div class="bx--time-picker {class}">
    <div class="bx--time-picker__input">{tmpl_element}</div>
    {tmpl_select_ampm}
    {tmpl_select_zone}
  </div>
  {tmpl_help}
</div>
"""
        return self.format(template, values, context)


    def render_tmpl_select_ampm(self, values, context):
        """Dynamically render a part of the component's template.
        """
        template = """
<div class="bx--time-picker__select bx--select {select_class}">
  <label for="select-ampm-{id}" class="bx--label bx--visually-hidden">
    {txt_select_ampm}
  </label>
  <select id="select-ampm-{id}" class="bx--select-input" {select_props}>
    <option class="bx--select-option" value="AM">AM</option>
    <option class="bx--select-option" value="PM">PM</option>
  </select>
  <svg focusable="false" preserveAspectRatio="xMidYMid meet"
      xmlns="http://www.w3.org/2000/svg" fill="currentColor"
      class="bx--select__arrow" width="16" height="16" viewBox="0 0 16 16"
      aria-hidden="true">
    <path d="M8 11L3 6 3.7 5.3 8 9.6 12.3 5.3 13 6z"></path>
  </svg>
</div>
"""
        return self.format(template, values)


    def render_tmpl_select_zone(self, values, context):
        """Dynamically render a part of the component's template.
        """
        template = """
<div class="bx--time-picker__select bx--select {select_class}">
  <label for="select-zone-{id}" class="bx--label bx--visually-hidden">
    {txt_select_zone}
  </label>
  <select id="select-zone-{id}" class="bx--select-input" {select_props}>
    {tmpl_timezones}
  </select>
  <svg focusable="false" preserveAspectRatio="xMidYMid meet"
      xmlns="http://www.w3.org/2000/svg" fill="currentColor"
      class="bx--select__arrow" width="16" height="16" viewBox="0 0 16 16"
      aria-hidden="true">
    <path d="M8 11L3 6 3.7 5.3 8 9.6 12.3 5.3 13 6z"></path>
  </svg>
</div>
"""
        return self.format(template, values, context)


    def render_tmpl_timezones(self, values, context):
        """Dynamically render a part of the component's template.
        """
        template = """
<option class="bx--select-option" value="{value}">{label}</option>
"""
        timezones = self.eval(self.kwargs.get('zones',
                pytz.common_timezones), context)
        items = []
        for zone in timezones:
            if isinstance(zone, str):
                value = label = zone
            else:
                value, label = zone
            items.append(template.format(value=value, label=label))
        return '\n'.join(items)


components = {
    'TimePicker': TimePicker,
}



"""
Breadcrumb
==========

See: https://www.carbondesignsystem.com/components/breadcrumb/usage/

The breadcrumb is a secondary navigation pattern that helps a user understand
the hierarchy among levels and navigate back through them.

Overview
--------

Breadcrumbs show users their current location relative to the information
architecture and enable them to quickly move up to a parent level or previous
step.
""" # pylint:disable=line-too-long
# pylint:disable=too-many-lines

from django.utils.translation import gettext as _
#-
from .base import Node

class Breadcrumb(Node):
    """Breadcrumb component.
    """
    WANT_CHILDREN = True
    "Template Tag needs closing end tag."
    NODE_PROPS = ('current',)
    "Extended Template Tag arguments."
    DEFAULT_TAG = 'nav'
    "Rendered HTML tag."

    def prepare(self, values, context):
        """Prepare values for rendering the templates.
        """
        values['txt_breadcrumb'] = _("breadcrumb")

        if self.eval(self.kwargs.get('current', False), context):
            values['class'].append('bx--breadcrumb--no-trailing-slash')


    def render_default(self, values, context):
        """Output html of the component.
        """
        template = """
<{tag} class="bx--breadcrumb {class}" aria-label="{txt_breadcrumb}" {props}>
  {child}
</{tag}>
"""
        return self.format(template, values)


class BreadcrumbItem(Node):
    """Breadcrumb item.
    """
    WANT_CHILDREN = True
    "Template Tag needs closing end tag."
    NODE_PROPS = ('href', 'current')
    "Extended Template Tag arguments."
    CLASS_AND_PROPS = ('wrapper',)
    "Prepare xxx_class and xxx_props values."

    def prepare(self, values, context):
        """Prepare values for rendering the templates.
        """
        values['href'] = self.eval(self.kwargs.get('href', '#'), context)

        if self.eval(self.kwargs.get('current', False), context):
            values['wrapper_class'].append('bx--breadcrumb-item--current')
            values['props'].append(('aria-current', 'page'))


    def render_default(self, values, context):
        """Output html of the component.
        """
        template = """
<{tag} class="bx--breadcrumb-item {wrapper_class}" {wrapper_props}>
  <a href="{href}" class="bx--link {class}" {props}>
    {child}
  </a>
</{tag}>
"""
        return self.format(template, values)


components = {
    'Breadcrumb': Breadcrumb,
    'BreadcrumbItem': BreadcrumbItem,
}



"""
Slider
======

See: https://www.carbondesignsystem.com/components/slider/usage/

Sliders provide a visual indication of adjustable content, where the user can
increase or decrease the value by moving the handle along a horizontal track.

Overview
--------

The slider in its basic form should be accompanied by a label and a number
input that doubles as a display for the slider’s current value.

The basic slider does not include discrete values, as the slider represents
a percentage of 0-100. In this case it is not necessary for a user to choose
a specific value, but instead generally increase or decrease an input. For
example, the user increases the slider amount and the volume of the music gets
louder.

The more complex versions should be used for selecting a specific value within
a value range.
""" # pylint:disable=line-too-long
# pylint:disable=too-many-lines

from django.utils.translation import gettext as _
#-
from .base import FormNode

class Slider(FormNode):
    """Slider component.
    """
    NODE_PROPS = ('id', 'min', 'max', 'step', 'light')
    "Extended Template Tag arguments."
    CLASS_AND_PROPS = ('label', 'slider')
    "Prepare xxx_class and xxx_props values."

    def prepare(self, values, context):
        """Prepare values for rendering the templates.
        """
        values['txt_slider'] = _("slider")
        values['min'] = self.eval(self.kwargs.get('min', 0), context)
        values['max'] = self.eval(self.kwargs.get('max', 100), context)
        values['step'] = self.eval(self.kwargs.get('step', 1), context)
        values['value'] = self.bound_field.value()

        if self.eval(self.kwargs.get('light'), context):
            values['class'].append('bx--text-input--light')

        if self.eval(self.kwargs.get('disabled'), context):
            values['label_class'].append('bx--label--disabled')
            values['slider_class'].append('bx--slider--disabled')


    def render_default(self, values, context):
        """Output html of the component.
        """
        template = """
<div class="bx--form-item">
  {tmpl_label}
  <div class="bx--slider-container">
    <label id="bottom_range-label-{id}" class="bx--slider__range-label">
      {min}
    </label>
    <div class="bx--slider {slider_class}" data-slider
        data-slider-input-box="#{id}" {slider_props}>
      <div class="bx--slider__thumb" tabindex="0"></div>
      <div class="bx--slider__track"></div>
      <div class="bx--slider__filled-track"></div>
      <input aria-label="{txt_slider}" id="slider-{id}"
          class="bx--slider__input" type="range" step="{step}" min="{min}"
          max="{max}" value="{value}">
    </div>
    <label id="top_range-label-{id}" class="bx--slider__range-label">
      {max}
    </label>
    <input id="{id}"
        aria-labelledby="bottom_range-label-{id} top_range-label-{id}"
        type="number" class="bx--text-input bx--slider-text-input {class}"
        placeholder="{min}" value="{value}" {props}>
  </div>
</div>
"""
        return self.format(template, values, context)


components = {
    'Slider': Slider,
}



"""
Notification
============

See: https://www.carbondesignsystem.com/components/notification/usage/

Notifications are messages that communicate information to the user. The two
main variants of notifications are toast notifications and inline notifications.

Overview
--------

Use notifications to inform users of updates or changes to system status.
Communicating with users and providing immediate feedback are important for
building trust. While notifications are an effective method of communicating
with users, they are disruptive and should be used sparingly.

For more context on when to use each notification variant, including modals,
refer to the notifications pattern. Carbon only supports inline, toast, and
modal notification variants, although some product teams also support banners
and notification centers.
""" # pylint:disable=line-too-long
# pylint:disable=too-many-lines

from django.contrib.messages import constants
from django.utils.translation import gettext as _
#-
from .base import Node

class Notification(Node):
    """Notification component.
    """
    WANT_CHILDREN = True
    "Template Tag needs closing end tag."
    SLOTS = ('action',)
    "Named children."
    MODES = ('inline', 'toast')
    "Available variants."
    NODE_PROPS = ('variant', 'low_contrast')
    "Extended Template Tag arguments."
    TEMPLATES = ('icon', 'close')
    "Conditional templates."
    POSSIBLE_VARIANT = ('info', 'success', 'warning', 'error')
    "Documentation only."

    variant = None

    def prepare(self, values, context):
        """Prepare values for rendering the templates.
        """
        values['txt_close'] = _("close")

        self.variant = self.eval(self.kwargs.get('variant', 'info'), context)
        if self.variant in (constants.DEBUG, constants.INFO):
            self.variant = 'info'
        elif self.variant == constants.SUCCESS:
            self.variant = 'success'
        elif self.variant == constants.WARNING:
            self.variant = 'warning'
        elif self.variant == constants.ERROR:
            self.variant = 'error'

        values['class'].append(f'bx--{self.mode}-notification--{self.variant}')

        if self.eval(self.kwargs.get('low_contrast'), context):
            values['class'].append(f'bx--{self.mode}-notification--low-contrast')

        context['mode'] = self.mode


    def render_inline(self, values, context):
        """Output html of the component.
        """
        template = """
<div data-notification class="bx--inline-notification {class}" role="alert">
  <div class="bx--inline-notification__details">
    {tmpl_icon}
    <div class="bx--inline-notification__text-wrapper">
      {child}
    </div>
  </div>
  {slot_action}
  {tmpl_close}
</div>
"""
        return self.format(template, values, context)


    def render_toast(self, values, context):
        """Output html of the component.
        """
        template = """
<div data-notification class="bx--toast-notification {class}" role="alert">
  {tmpl_icon}
  <div class="bx--toast-notification__details">
    {child}
  </div>
  {tmpl_close}
</div>
"""
        return self.format(template, values, context)


    def render_tmpl_icon(self, values, context):
        """Dynamically render a part of the component's template.
        """
        if self.variant == 'info':
            template = """
<svg focusable="false" preserveAspectRatio="xMidYMid meet"
    xmlns="http://www.w3.org/2000/svg" fill="currentColor"
    class="bx--{mode}-notification__icon" width="20" height="20"
    viewBox="0 0 32 32" aria-hidden="true">
  <path fill="none" d="M16,8a1.5,1.5,0,1,1-1.5,1.5A1.5,1.5,0,0,1,16,8Zm4,13.875H17.125v-8H13v2.25h1.875v5.75H12v2.25h8Z" data-icon-path="inner-path"></path>
  <path d="M16,2A14,14,0,1,0,30,16,14,14,0,0,0,16,2Zm0,6a1.5,1.5,0,1,1-1.5,1.5A1.5,1.5,0,0,1,16,8Zm4,16.125H12v-2.25h2.875v-5.75H13v-2.25h4.125v8H20Z"></path>
</svg>
"""
        elif self.variant == 'error':
            template = """
<svg focusable="false" preserveAspectRatio="xMidYMid meet"
    xmlns="http://www.w3.org/2000/svg" fill="currentColor"
    class="bx--{mode}-notification__icon" width="20" height="20"
    viewBox="0 0 20 20" aria-hidden="true">
  <path d="M10,1c-5,0-9,4-9,9s4,9,9,9s9-4,9-9S15,1,10,1z M13.5,14.5l-8-8l1-1l8,8L13.5,14.5z"></path>
  <path d="M13.5,14.5l-8-8l1-1l8,8L13.5,14.5z" data-icon-path="inner-path" opacity="0"></path>
</svg>
"""
        elif self.variant == 'success':
            template = """
<svg focusable="false" preserveAspectRatio="xMidYMid meet"
    xmlns="http://www.w3.org/2000/svg" fill="currentColor"
    class="bx--{mode}-notification__icon" width="20" height="20"
    viewBox="0 0 20 20" aria-hidden="true">
  <path d="M10,1c-4.9,0-9,4.1-9,9s4.1,9,9,9s9-4,9-9S15,1,10,1z M8.7,13.5l-3.2-3.2l1-1l2.2,2.2l4.8-4.8l1,1L8.7,13.5z"></path>
  <path fill="none" d="M8.7,13.5l-3.2-3.2l1-1l2.2,2.2l4.8-4.8l1,1L8.7,13.5z" data-icon-path="inner-path" opacity="0"></path>
</svg>
"""
        elif self.variant == 'warning':
            template = """
<svg focusable="false" preserveAspectRatio="xMidYMid meet"
    xmlns="http://www.w3.org/2000/svg" fill="currentColor"
    class="bx--{mode}-notification__icon" width="20" height="20"
    viewBox="0 0 20 20" aria-hidden="true">
  <path d="M10,1c-5,0-9,4-9,9s4,9,9,9s9-4,9-9S15,1,10,1z M9.2,5h1.5v7H9.2V5z M10,16c-0.6,0-1-0.4-1-1s0.4-1,1-1  s1,0.4,1,1S10.6,16,10,16z"></path>
  <path d="M9.2,5h1.5v7H9.2V5z M10,16c-0.6,0-1-0.4-1-1s0.4-1,1-1s1,0.4,1,1S10.6,16,10,16z" data-icon-path="inner-path" opacity="0"></path>
</svg>
"""
        return self.format(template, values)


    def render_tmpl_close(self, values, context):
        """Dynamically render a part of the component's template.
        """
        template = """
<button data-notification-btn class="bx--{mode}-notification__close-button"
    type="button" aria-label="{txt_close}">
  <svg focusable="false" preserveAspectRatio="xMidYMid meet"
      xmlns="http://www.w3.org/2000/svg" fill="currentColor"
      class="bx--{mode}-notification__close-icon" width="20" height="20"
      viewBox="0 0 32 32" aria-hidden="true">
    <path d="M24 9.4L22.6 8 16 14.6 9.4 8 8 9.4 14.6 16 8 22.6 9.4 24 16 17.4 22.6 24 24 22.6 17.4 16 24 9.4z"></path>
  </svg>
</button>
"""
        return self.format(template, values)


class NotificationButton(Node):
    """Notification button component.
    """
    WANT_CHILDREN = True
    "Template Tag needs closing end tag."
    DEFAULT_TAG = 'button'
    "Rendered HTML tag."

    def render_default(self, values, context):
        """Output html of the component.
        """
        template = """
<{tag} tabindex="0"
    class="bx--inline-notification__action-button bx--btn bx--btn--sm bx--btn--ghost {class}"
    {props}>
  {child}
</{tag}>
"""
        return self.format(template, values)


class NotificationTitle(Node):
    """Notification title component.
    """
    WANT_CHILDREN = True
    "Template Tag needs closing end tag."
    DEFAULT_TAG = 'h3'
    "Rendered HTML tag."

    def render_default(self, values, context):
        """Output html of the component.
        """
        values['mode'] = context.get('mode', 'inline')
        template = """
<{tag} class="bx--{mode}-notification__title {class}" {props}>{child}</{tag}>
"""
        return self.format(template, values)


class NotificationSubtitle(Node):
    """Notification subtitle component.
    """
    WANT_CHILDREN = True
    "Template Tag needs closing end tag."
    DEFAULT_TAG = 'p'
    "Rendered HTML tag."

    def render_default(self, values, context):
        """Output html of the component.
        """
        values['mode'] = context.get('mode', 'inline')
        template = """
<{tag} class="bx--{mode}-notification__subtitle {class}" {props}>{child}</{tag}>
"""
        return self.format(template, values)


class NotificationCaption(Node):
    """Notification caption component.
    """
    WANT_CHILDREN = True
    "Template Tag needs closing end tag."
    DEFAULT_TAG = 'p'
    "Rendered HTML tag."

    def render_default(self, values, context):
        """Output html of the component.
        """
        values['mode'] = context.get('mode', 'inline')
        template = """
<{tag} class="bx--{mode}-notification__caption {class}" {props}>{child}</{tag}>
"""
        return self.format(template, values)


components = {
    'Notification': Notification,
    'NotificationButton': NotificationButton,
    'NotificationTitle': NotificationTitle,
    'NotificationSubtitle': NotificationSubtitle,
    'NotificationCaption': NotificationCaption,
}



"""
Data Table
==========

See: https://www.carbondesignsystem.com/components/data-table/usage/

Data tables are used to organize and display data efficiently. The data table
component allows for customization with additional functionality, as needed by
your product’s users.

Overview
--------

The data table’s features are ideal for organizing and displaying data in a UI.
The column headers can sort data in ascending or descending order, rows can be
expanded to progressively disclose information, and single or batch actions can
be taken on rows.

The data table toolbar gives a location for primary buttons, search, filtering,
table display settings, and other utilities.
""" # pylint:disable=line-too-long
# pylint:disable=too-many-lines

import logging
from uuid import uuid4
#-
from django.utils.translation import gettext as _
#-
from .base import DummyNodeList, FormNode, Node, clean_attr_value, modify_svg
from .button import Button

_logger = logging.getLogger(__name__)


class DataTable(Node):
    """Data Table component.
    """
    WANT_CHILDREN = True
    "Template Tag needs closing end tag."
    SLOTS = ('title', 'description', 'batch_actions', 'search',
            'toolbar_actions', 'toolbar_overflow', 'head', 'foot', 'pagination')
    "Named children."
    MODES = ('default', 'sticky')
    "Available variants."
    NODE_PROPS = ('variant', 'sortable', 'small_toolbar', 'batch_field',
            'visible_overflow')
    "Extended Template Tag arguments."
    CLASS_AND_PROPS = ('toolbar', *Node.CLASS_AND_PROPS)
    "Prepare xxx_class and xxx_props values."
    POSSIBLE_VARIANT = ('compact', 'short', 'tall', 'zebra')
    "Documentation only."
    PAGER_SIZES = (10, 20, 30, 40, 50)

    def prepare(self, values, context):
        """Prepare values for rendering the templates.
        """
        values['txt_batch_actions'] = _("Table Action Bar")
        values['txt_items_selected'] = _("items selected")
        values['txt_overflow'] = _("Overflow")

        variant = self.eval(self.kwargs.get('variant'), context)
        if variant:
            values['class'].append(f'bx--data-table--{variant}')
        if variant == 'compact':
            context['compact'] = True

        if self.eval(self.kwargs.get('sortable'), context):
            values['class'].append('bx--data-table--sort')

        if variant == 'compact' or\
                self.eval(self.kwargs.get('small_toolbar'), context):
            values['toolbar_class'].append('bx--table-toolbar--small')

            self.set_child_props(context, 'button_props', 'batch_actions',
                    small=True)
            self.set_child_props(context, 'search_props', small=True)
            self.set_child_props(context, 'button_props', 'toolbar',
                    small=True)

        if self.eval(self.kwargs.get('visible_overflow'), context):
            values['class'].append('bx--data-table--visible-overflow-menu')


    def render_default(self, values, context):
        """Output html of the component.
        """
        if self.has_tmpl_header() or self.has_tmpl_toolbar():
            template = """
<div class="bx--data-table-container" data-table>
  {tmpl_header}
  {tmpl_toolbar}
  <table class="bx--data-table {class}" {props}>
    {slot_head}
    {child}
    {slot_foot}
  </table>
  {slot_pagination}
</div>
"""
        else:
            template = """
<table class="bx--data-table {class}" {props}>
  {slot_head}
  {child}
  {slot_foot}
</table>
{slot_pagination}
"""
        return self.format(template, values, context)


    def render_sticky(self, values, context):
        """Output html of the component.
        """
        if self.has_tmpl_header() or self.has_tmpl_toolbar():
            template = """
<div class="bx--data-table-container" data-table>
  {tmpl_header}
  {tmpl_toolbar}
  <section class="bx--data-table_inner-container">
    <table class="bx--data-table bx--data-table--sticky-header {class}" {props}>
      {slot_head}
      {child}
      {slot_foot}
    </table>
  </section>
  {slot_pagination}
</div>
"""
        else:
            template = """
<section class="bx--data-table_inner-container">
  <table class="bx--data-table bx--data-table--sticky-header {class}" {props}>
    {slot_head}
    {child}
    {slot_foot}
  </table>
</section>
{slot_pagination}
"""
        return self.format(template, values, context)


    def render_slot_head(self, values, context):
        """Render html of the slot.
        """
        template = '<thead class="{class}" {props}>{child}</thead>'
        return self.format(template, values)


    def render_slot_foot(self, values, context):
        """Render html of the slot.
        """
        template = '<tfoot class="{class}" {props}>{child}</tfoot>'
        return self.format(template, values)


    def render_slot_title(self, values, context):
        """Render html of the slot.
        """
        template = """
<h4 class="bx--data-table-header__title {class}" {props}>
  {child}
</h4>
"""
        return self.format(template, values)


    def render_slot_description(self, values, context):
        """Render html of the slot.
        """
        template = """
<p class="bx--data-table-header__description {class}" {props}>
  {child}
</p>
"""
        return self.format(template, values)


    def render_slot_batch_actions(self, values, context):
        """Render html of the slot.
        """
        cancel_btn = Button(
                DummyNodeList(_("Cancel")),
                **{
                    'variant': 'primary',
                    'data-event': 'action-bar-cancel',
                    'class': 'bx--batch-summary__cancel',
                    'small': context.get('compact'),
                })

        form_field = self.eval(self.kwargs.get('batch_field'), context)
        if form_field and form_field.value():
            values['count'] = len(form_field.value())
        else:
            values['count'] = 0

        template = """
<div class="bx--batch-actions {class}" aria-label="{txt_batch_actions}" {props}>
  <div class="bx--action-list">
    {child}
    {cancel_btn}
  </div>
  <div class="bx--batch-summary">
    <p class="bx--batch-summary__para">
      <span data-items-selected>{count}</span> {txt_items_selected}
    </p>
  </div>
</div>
"""
        return self.format(template,
                dict(values, cancel_btn=cancel_btn.render(context)))


    def render_slot_toolbar_overflow(self, values, context):
        """Render html of the slot.
        """
        template = """
<div class="bx--overflow-menu bx--toolbar-action {class}" data-overflow-menu
    role="button" tabindex="0" aria-label="{txt_overflow}" aria-haspopup="true"
    aria-expanded="false" {props}>
  <svg focusable="false" preserveAspectRatio="xMidYMid meet"
      xmlns="http://www.w3.org/2000/svg" fill="currentColor"
      class="bx--toolbar-action__icon" width="16" height="16"
      viewBox="0 0 16 16" aria-hidden="true">
    <path d="M13.5,8.4c0-0.1,0-0.3,0-0.4c0-0.1,0-0.3,0-0.4l1-0.8c0.4-0.3,0.4-0.9,0.2-1.3l-1.2-2C13.3,3.2,13,3,12.6,3	c-0.1,0-0.2,0-0.3,0.1l-1.2,0.4c-0.2-0.1-0.4-0.3-0.7-0.4l-0.3-1.3C10.1,1.3,9.7,1,9.2,1H6.8c-0.5,0-0.9,0.3-1,0.8L5.6,3.1	C5.3,3.2,5.1,3.3,4.9,3.4L3.7,3C3.6,3,3.5,3,3.4,3C3,3,2.7,3.2,2.5,3.5l-1.2,2C1.1,5.9,1.2,6.4,1.6,6.8l0.9,0.9c0,0.1,0,0.3,0,0.4	c0,0.1,0,0.3,0,0.4L1.6,9.2c-0.4,0.3-0.5,0.9-0.2,1.3l1.2,2C2.7,12.8,3,13,3.4,13c0.1,0,0.2,0,0.3-0.1l1.2-0.4	c0.2,0.1,0.4,0.3,0.7,0.4l0.3,1.3c0.1,0.5,0.5,0.8,1,0.8h2.4c0.5,0,0.9-0.3,1-0.8l0.3-1.3c0.2-0.1,0.4-0.2,0.7-0.4l1.2,0.4	c0.1,0,0.2,0.1,0.3,0.1c0.4,0,0.7-0.2,0.9-0.5l1.1-2c0.2-0.4,0.2-0.9-0.2-1.3L13.5,8.4z M12.6,12l-1.7-0.6c-0.4,0.3-0.9,0.6-1.4,0.8	L9.2,14H6.8l-0.4-1.8c-0.5-0.2-0.9-0.5-1.4-0.8L3.4,12l-1.2-2l1.4-1.2c-0.1-0.5-0.1-1.1,0-1.6L2.2,6l1.2-2l1.7,0.6	C5.5,4.2,6,4,6.5,3.8L6.8,2h2.4l0.4,1.8c0.5,0.2,0.9,0.5,1.4,0.8L12.6,4l1.2,2l-1.4,1.2c0.1,0.5,0.1,1.1,0,1.6l1.4,1.2L12.6,12z"></path>
    <path d="M8,11c-1.7,0-3-1.3-3-3s1.3-3,3-3s3,1.3,3,3C11,9.6,9.7,11,8,11C8,11,8,11,8,11z M8,6C6.9,6,6,6.8,6,7.9C6,7.9,6,8,6,8	c0,1.1,0.8,2,1.9,2c0,0,0.1,0,0.1,0c1.1,0,2-0.8,2-1.9c0,0,0-0.1,0-0.1C10,6.9,9.2,6,8,6C8.1,6,8,6,8,6z"></path>
  </svg>
  <ul class="bx--overflow-menu-options bx--overflow-menu--flip" tabindex="-1"
      role="menu" aria-label="{txt_overflow}" data-floating-menu-direction="bottom">
    {child}
  </ul>
</div>
"""
        return self.format(template, values)


    def has_tmpl_header(self):
        """Should render header template?
        """
        return 'title' in self.slots or 'description' in self.slots

    def render_tmpl_header(self, values, context):
        """Dynamically render a part of the component's template.
        """
        if not self.has_tmpl_header():
            return ''

        template = """
<div class="bx--data-table-header">
  {slot_title}
  {slot_description}
</div>
"""
        return self.format(template, values, context)


    def has_tmpl_toolbar(self):
        """Should render toolbar template?
        """
        return 'batch_actions' in self.slots or\
                'toolbar_overflow' in self.slots or\
                'toolbar_actions' in self.slots

    def render_tmpl_toolbar(self, values, context):
        """Dynamically render a part of the component's template.
        """
        if not self.has_tmpl_toolbar():
            return ''

        template = """
<section class="bx--table-toolbar {toolbar_class}" {toolbar_props}>
  {slot_batch_actions}
  <div class="bx--toolbar-content">
    {slot_search}
    {slot_toolbar_overflow}
    {slot_toolbar_actions}
  </div>
</section>
"""
        return self.format(template, values, context)


class TrExpandable(Node):
    """Table row for expandable table.
    """
    WANT_CHILDREN = True
    "Template Tag needs closing end tag."
    SLOTS = ('subrow',)
    "Named children."

    def render_default(self, values, context):
        """Output html of the component.
        """
        values['total_column'] = values['child'].count('<td') + 1

        template = """
<tr class="bx--parent-row {class}" data-parent-row {props}>
  <td class="bx--table-expand" data-event="expand">
    <button class="bx--table-expand__button">
      <svg focusable="false" preserveAspectRatio="xMidYMid meet"
          xmlns="http://www.w3.org/2000/svg" fill="currentColor"
          class="bx--table-expand__svg" width="16" height="16"
          viewBox="0 0 16 16" aria-hidden="true">
        <path d="M11 8L6 13 5.3 12.3 9.6 8 5.3 3.7 6 3z"></path>
      </svg>
    </button>
  </td>
  {child}
</tr>
<tr class="bx--expandable-row bx--expandable-row--hidden" data-child-row>
  <td colspan="{total_column}">
    <div class="bx--child-row-inner-container">
      {slot_subrow}
    </div>
  </td>
</tr>
"""
        return self.format(template, values, context)


class Th(Node):
    """Table row heading.
    """
    WANT_CHILDREN = True
    "Template Tag needs closing end tag."
    MODES = ('default', 'checkbox', 'sortable', 'menu', 'expandable',
            'expand_all', 'row')
    "Available variants."
    NODE_PROPS = ('id',)
    "Extended Template Tag arguments."

    def render_default(self, values, context):
        """Output html of the component.
        """
        template = """
<th class="{class}" {props}>
  <span class="bx--table-header-label">{child}</span>
</th>
"""
        return self.format(template, values)


    def render_checkbox(self, values, context):
        """Output html of the component.
        """
        template = """
<th class="bx--table-column-checkbox {class}">
  <input type="checkbox" id="{id}" data-event="select-all" class="bx--checkbox" {props}>
  <label for="{id}" class="bx--checkbox-label" aria-label="{label}"></label>
</th>
"""
        return self.format(template, values)


    def render_sortable(self, values, context):
        """Output html of the component.
        """
        values['cleaned_child'] = clean_attr_value(values['child'])

        template = """
<th class="{class}" {props}>
  <button class="bx--table-sort" data-event="sort" title="{cleaned_child}">
    <span class="bx--table-header-label">{child}</span>
    <svg focusable="false" preserveAspectRatio="xMidYMid meet"
        xmlns="http://www.w3.org/2000/svg" fill="currentColor"
        class="bx--table-sort__icon" width="16" height="16"
        viewBox="0 0 16 16" aria-hidden="true">
      <path d="M12.3 9.3L8.5 13.1 8.5 1 7.5 1 7.5 13.1 3.7 9.3 3 10 8 15 13 10z"></path>
    </svg>
    <svg focusable="false" preserveAspectRatio="xMidYMid meet"
        xmlns="http://www.w3.org/2000/svg" fill="currentColor"
        class="bx--table-sort__icon-unsorted" width="16" height="16"
        viewBox="0 0 32 32" aria-hidden="true">
      <path d="M27.6 20.6L24 24.2 24 4 22 4 22 24.2 18.4 20.6 17 22 23 28 29 22zM9 4L3 10 4.4 11.4 8 7.8 8 28 10 28 10 7.8 13.6 11.4 15 10z"></path>
    </svg>
  </button>
</th>
"""
        return self.format(template, values)


    def render_menu(self, values, context):
        """Output html of the component.
        """
        return '<th class="bx--table-column-menu"></th>'


    def render_expandable(self, values, context):
        """Output html of the component.
        """
        template = """
<th class="bx--table-expand {class}" data-event="expandAll" {props}>
  <span class="bx--table-header-label"></span>
</th>
"""
        return self.format(template, values)


    def render_expand_all(self, values, context):
        """Output html of the component.
        """
        template = """
<th class="bx--table-expand {class}" data-event="expandAll" {props}>
  <button class="bx--table-expand__button">
    <svg focusable="false" preserveAspectRatio="xMidYMid meet"
        xmlns="http://www.w3.org/2000/svg" fill="currentColor"
        class="bx--table-expand__svg" width="16" height="16" viewBox="0 0 16 16"
        aria-hidden="true">
      <path d="M11 8L6 13 5.3 12.3 9.6 8 5.3 3.7 6 3z"></path>
    </svg>
  </button>
</th>
"""
        return self.format(template, values)


    def render_row(self, values, context):
        """Output html of the component.
        """
        template = '<th scope="row" class="{class}" {props}>{child}</th>'
        return self.format(template, values)


class Td(Node):
    """Table row column.
    """
    WANT_CHILDREN = True
    "Template Tag needs closing end tag."
    MODES = ('default', 'menu', 'menu_visible')
    "Available variants."
    SLOTS = ('secondary',)
    "Named children."

    def prepare(self, values, context):
        """Prepare values for rendering the templates.
        """
        values['txt_menu'] = _("Open menu")


    def render_default(self, values, context):
        """Output html of the component.
        """
        template = """
<td class="{class}" {props}>
  {child}
  {slot_secondary}
</td>
"""
        return self.format(template, values, context)


    def render_menu(self, values, context):
        """Output html of the component.
        """
        template = """
<td class="bx--table-column-menu {class}" {props}>
  <div data-overflow-menu role="menu" tabindex="0" aria-label="{label}"
      class="bx--overflow-menu" title="{txt_menu}">
    <svg focusable="false" preserveAspectRatio="xMidYMid meet"
        xmlns="http://www.w3.org/2000/svg" fill="currentColor"
        class="bx--overflow-menu__icon" width="16" height="16"
        viewBox="0 0 32 32" aria-hidden="true">
      <circle cx="16" cy="8" r="2"></circle>
      <circle cx="16" cy="16" r="2"></circle>
      <circle cx="16" cy="24" r="2"></circle>
    </svg>
    <ul class="bx--overflow-menu-options bx--overflow-menu--flip">
      {child}
    </ul>
  </div>
</td>
"""
        return self.format(template, values)


    def render_menu_visible(self, values, context):
        """Output html of the component.
        """
        template = """
<td class="bx--table-column-menu {class}" {props}>
  <div data-overflow-menu role="menu" tabindex="0"
      aria-label="{label}" class="bx--overflow-menu" title="{txt_menu}">
    <svg focusable="false" preserveAspectRatio="xMidYMid meet"
        fill="currentColor" xmlns="http://www.w3.org/2000/svg"
        class="bx--overflow-menu__icon" width="16" height="16"
        viewBox="0 0 16 16" aria-hidden="true">
      <circle cx="8" cy="3" r="1"></circle>
      <circle cx="8" cy="8" r="1"></circle>
      <circle cx="8" cy="13" r="1"></circle>
    </svg>
    <ul class="bx--overflow-menu-options bx--overflow-menu--flip">
      {child}
    </ul>
  </div>
</td>
"""
        return self.format(template, values)


    def render_slot_secondary(self, values, context):
        """Render html of the slot.
        """
        template = """
<div class="bx--data-table--cell-secondary-text {class}" {props}>
  {child}
</div>
"""
        return self.format(template, values)


class TdCheckBox(FormNode):
    """Table row checkbox.
    """
    NODE_PROPS = ('value',)
    "Extended Template Tag arguments."
    REQUIRED_PROPS = ('value',)
    "Will raise Exception if not set."

    value = None

    def default_id(self):
        """Get Django form field html id.
        """
        return '%s-%s' % (self.bound_field.id_for_label, uuid4().hex)


    def label(self):
        """Get Django form field label.
        """
        for _, val, txt in self.choices():
            if val == self.value:
                return txt
        return ''


    def before_prepare(self, values, context):
        """Initialize the values meant for rendering templates.
        """
        self.value = self.eval(self.kwargs['value'], context)
        super().before_prepare(values, context)


    def prepare(self, values, context):
        """Prepare values for rendering the templates.
        """
        values['value'] = self.value
        values['name'] = self.bound_field.name

        if self.bound_value and self.value in self.bound_value:
            values['props'].append(('checked', True))


    def render_default(self, values, context):
        """Output html of the component.
        """
        template = """
<td class="bx--table-column-checkbox">
  <input name="{name}" value="{value}" type="checkbox" id="{id}"
      class="bx--checkbox {class}" data-event="select" {props}>
  <label for="{id}" class="bx--checkbox-label" aria-label="{label}"></label>
</td>
"""
        return self.format(template, values)


class TableToolbarSearch(Node):
    """Table toolbar search filter.
    """
    NODE_PROPS = ('id', 'expandable', 'small')
    "Extended Template Tag arguments."
    CLASS_AND_PROPS = ('wrapper', 'magnifier')
    "Prepare xxx_class and xxx_props values."

    CATCH_PROPS = ('search_props',)

    def prepare(self, values, context):
        """Prepare values for rendering the templates.
        """
        values['txt_search'] = _("Search")
        values['txt_clear'] = _("Clear search input")
        if not values['label']:
            values['label'] = values['txt_search']

        if self.eval(self.kwargs.get('expandable'), context):
            values['wrapper_class'].append(
                    'bx--toolbar-search-container-expandable')
            values['magnifier_props'].append(('tabindex', '0'))
        else:
            values['wrapper_class'].append(
                    'bx--toolbar-search-container-persistent')

        if self.eval(self.kwargs.get('small'), context):
            values['class'].append('bx--search--sm')


    def render_default(self, values, context):
        """Output html of the component.
        """
        template = """
<div class="{wrapper_class}" {wrapper_props}>
  <div data-search class="bx--search {class}" role="search" {props}>
    <div class="bx--search-magnifier {magnifier_class}" {magnifier_props}>
      {tmpl_magnifier_icon}
    </div>
    <label id="label-{id}" class="bx--label" for="{id}">
      {label}
    </label>
    <input class="bx--search-input" type="text" id="{id}" role="search"
        placeholder="{txt_search}" aria-labelledby="label-{id}">
    <button class="bx--search-close bx--search-close--hidden"
        title="{txt_clear}" aria-label="{txt_clear}">
      <svg focusable="false" preserveAspectRatio="xMidYMid meet"
          xmlns="http://www.w3.org/2000/svg" fill="currentColor"
          width="16" height="16" viewBox="0 0 32 32" aria-hidden="true">
        <path d="M24 9.4L22.6 8 16 14.6 9.4 8 8 9.4 14.6 16 8 22.6 9.4 24 16 17.4 22.6 24 24 22.6 17.4 16 24 9.4z"></path>
      </svg>
    </button>
  </div>
</div>
"""
        return self.format(template, values, context)


    def render_tmpl_magnifier_icon(self, values, context):
        """Dynamically render a part of the component's template.
        """
        return """
<svg focusable="false" preserveAspectRatio="xMidYMid meet"
    xmlns="http://www.w3.org/2000/svg" fill="currentColor"
    class="bx--toolbar-action__icon" width="16" height="16" viewBox="0 0 16 16"
    aria-hidden="true">
  <path d="M15,14.3L10.7,10c1.9-2.3,1.6-5.8-0.7-7.7S4.2,0.7,2.3,3S0.7,8.8,3,10.7c2,1.7,5,1.7,7,0l4.3,4.3L15,14.3z M2,6.5	C2,4,4,2,6.5,2S11,4,11,6.5S9,11,6.5,11S2,9,2,6.5z"></path>
</svg>
"""


class TableToolbarOverflowButton(Node):
    """Table toolbar action button.
    """
    WANT_CHILDREN = True
    "Template Tag needs closing end tag."
    NODE_PROPS = ('active',)
    "Extended Template Tag arguments."
    CLASS_AND_PROPS = ('list',)
    "Prepare xxx_class and xxx_props values."

    def prepare(self, values, context):
        """Prepare values for rendering the templates.
        """
        if self.eval(self.kwargs.get('active'), context):
            values['props'].append(('data-floating-menu-primary-focus', True))

        if not context.get('compact'):
            values['list_class'].append('bx--overflow-menu--data-table')


    def render_default(self, values, context):
        """Output html of the component.
        """
        template = """
<li class="bx--overflow-menu-options__option {list_class}"
    role="presentation" {list_props}>
  <button class="bx--overflow-menu-options__btn {class}" role="menuitem"
      {props}>
    <div class="bx--overflow-menu-options__option-content">
      {child}
    </div>
  </button>
</li>
"""
        return self.format(template, values)


class TdOverflowButton(Node):
    """Table row action button.
    """
    WANT_CHILDREN = True
    "Template Tag needs closing end tag."
    SLOTS = ('icon',)
    "Named children."
    NODE_PROPS = ('icon_size',)
    "Extended Template Tag arguments."

    def render_default(self, values, context):
        """Output html of the component.
        """
        values['cleaned_child'] = clean_attr_value(values['child'])

        template = """
<li class="bx--overflow-menu-options__option bx--table-row--menu-option">
  <{tag} class="bx--overflow-menu-options__btn {class}"
      title="{cleaned_child}" {props}>
    <div class="bx--overflow-menu-options__option-content">
      {slot_icon}
      {child}
    </div>
  </{tag}>
</li>
"""
        return self.format(template, values, context)


    def render_slot_icon(self, values, context):
        """Render html of the slot.
        """
        size = self.eval(self.kwargs.get('icon_size', 16), context)
        return modify_svg(values['child'], {
            'focusable': 'false',
            'preserveAspectRatio': 'xMidYMid meet',
            'fill': 'currentColor',
            'style': {
                'width': '%spx' % size,
                'height': '%spx' % size,
            },
            'aria-hidden': 'true',
            'class': values['class'],
        })


components = {
    'Table': DataTable,
    'Th': Th,
    'Td': Td,
    'TdCheck': TdCheckBox,
    'TableOvButton': TableToolbarOverflowButton,
    'TdOvButton': TdOverflowButton,
    'TbSearch': TableToolbarSearch,
    'TrExpandable': TrExpandable,
}



"""
Accordion
=========

See: https://www.carbondesignsystem.com/components/accordion/usage/

An accordion is a vertically stacked list of headers that reveal or hide
associated sections of content.

Overview
--------

The accordion component delivers large amounts of content in a small space
through progressive disclosure. The header title give the user a high level
overview of the content allowing the user to decide which sections to read.

Accordions can make information processing and discovering more effective.
However, it does hide content from users and it’s important to account for
a user not noticing or reading all of the included content. If a user is
likely to read all of the content then don’t use an accordion as it adds
the burden of an extra click; instead use a full scrolling page with normal
headers.
""" # pylint:disable=line-too-long
# pylint:disable=too-many-lines

from .base import Node

class Accordion(Node):
    """Accordion component.
    """
    WANT_CHILDREN = True
    "Template Tag needs closing end tag."
    DEFAULT_TAG = 'ul'
    "Rendered HTML tag."

    def render_default(self, values, context):
        """Output html of the component.
        """
        template = """
<{tag} data-accordion class="bx--accordion {class}" {props}>
  {child}
</{tag}>
"""
        return self.format(template, values)


class AccordionItem(Node):
    """Accordion item.
    """
    WANT_CHILDREN = True
    "Template Tag needs closing end tag."
    NODE_PROPS = ('id', 'expanded')
    "Extended Template Tag arguments."
    DEFAULT_TAG = 'li'
    "Rendered HTML tag."

    def prepare(self, values, context):
        """Prepare values for rendering the templates.
        """
        if self.eval(self.kwargs.get('expanded', False), context):
            values['expanded'] = 'true'
            values['class'].append('bx--accordion__item--active')
        else:
            values['expanded'] = 'false'


    def render_default(self, values, context):
        """Output html of the component.
        """
        template = """
<{tag} data-accordion-item class="bx--accordion__item {class}" {props}>
  <button class="bx--accordion__heading" aria-expanded="{expanded}"
      aria-controls="{id}">
    <svg focusable="false" preserveAspectRatio="xMidYMid meet"
        xmlns="http://www.w3.org/2000/svg" fill="currentColor"
        class="bx--accordion__arrow" width="16" height="16" viewBox="0 0 16 16"
        aria-hidden="true">
      <path d="M11 8L6 13 5.3 12.3 9.6 8 5.3 3.7 6 3z"></path>
    </svg>
    {tmpl_label}
  </button>
  <div id="{id}" class="bx--accordion__content">
    {child}
  </div>
</{tag}>
"""
        return self.format(template, values, context)


    def render_tmpl_label(self, values, context):
        """Dynamically render a part of the component's template.
        """
        if not values['label']:
            return ''
        template = """
<div class="bx--accordion__title">
  {label}
</div>
"""
        return self.format(template, values)


components = {
    'Accordion': Accordion,
    'AccordionItem': AccordionItem,
}



"""
Tag
===

See: https://www.carbondesignsystem.com/components/tag/usage/

Use tags to label, categorize, or organize items using keywords that describe
them.

Overview
--------

Multiple or single tags can be used to categorize items.

Use short labels for easy scanning. Use two words only if necessary to
describe the status and differentiate it from other tags.
""" # pylint:disable=line-too-long
# pylint:disable=too-many-lines

from django.utils.translation import gettext as _
#-
from .base import Node

class Tag(Node):
    """Tag component.
    """
    WANT_CHILDREN = True
    "Template Tag needs closing end tag."
    MODES = ('default', 'filter')
    "Available variants."
    NODE_PROPS = ('variant',)
    "Extended Template Tag arguments."

    def prepare(self, values, context):
        """Prepare values for rendering the templates.
        """
        values['txt_clear'] = _("Clear filter")

        variant = self.eval(self.kwargs.get('variant'), context)
        if variant:
            values['class'].append(f'bx--tag--{variant}')


    def render_default(self, values, context):
        """Output html of the component.
        """
        template = """
<button class="bx--tag {class}" {props}>
  <span class="bx--tag__label">{child}</span>
</button>
"""
        return self.format(template, values)


    def render_filter(self, values, context):
        """Output html of the component.
        """
        template = """
<div class="bx--tag bx--tag--filter" title="{txt_clear}">
  <span class="bx--tag__label">{child}</span>
  <button class="bx--tag__close-icon">
    <svg focusable="false" preserveAspectRatio="xMidYMid meet"
        xmlns="http://www.w3.org/2000/svg" fill="currentColor" width="16"
        height="16" viewBox="0 0 32 32" aria-hidden="true">
      <path d="M24 9.4L22.6 8 16 14.6 9.4 8 8 9.4 14.6 16 8 22.6 9.4 24 16 17.4 22.6 24 24 22.6 17.4 16 24 9.4z"></path>
    </svg>
  </button>
</div>
"""
        return self.format(template, values)


components = {
    'Tag': Tag,
}



"""
Tile
====

See: https://www.carbondesignsystem.com/components/tile/usage/

Tiles are a highly flexible component for displaying a wide variety of
content, including information, getting started, how-to, next steps, and more.

Overview
--------

Carbon ships a basic tile structure that responds to the grid. Based on the
layout structure, tiles can contain type, images and/or a block of color.
As tiles have no pre-set styles for the content within them, the tile
component usage guidance is purposefully high-level. It focuses specifically
on the tile itself, not the structure of the information or interactive
elements that the tile contains—that type of information will be found in
the card pattern.
""" # pylint:disable=line-too-long
# pylint:disable=too-many-lines

from django.utils.translation import gettext as _
#-
from .base import ChoiceFormNode, Node

class Tile(Node):
    """Tile component.
    """
    WANT_CHILDREN = True
    "Template Tag needs closing end tag."
    SLOTS = ('above', 'icon')
    "Named children."
    MODES = ('default', 'clickable', 'expandable')
    "Available variants."

    def prepare(self, values, context):
        """Prepare values for rendering the templates.
        """
        values['txt_tile'] = _("tile")

        if self.mode == 'selectable':
            label = self.eval(self.kwargs.get('label'), context)
            if label:
                values['props'].append(('title', label))
                values['label_props'].append(('aria-label', label))


    def render_default(self, values, context):
        """Output html of the component.
        """
        template = """
<div class="bx--tile {class}" {props}>
  {child}
</div>
"""
        return self.format(template, values)


    def render_clickable(self, values, context):
        """Output html of the component.
        """
        template = """
<a data-tile="clickable" class="bx--tile bx--tile--clickable {class}"
    tabindex="0" {props}>
  {child}
</a>
"""
        return self.format(template, values)


    def render_expandable(self, values, context):
        """Output html of the component.
        """
        template = """
<div data-tile="expandable" class="bx--tile bx--tile--expandable {class}"
    tabindex="0" {props}>
  <button aria-label="expand menu" class="bx--tile__chevron">
    {tmpl_icon}
  </button>
  <div class="bx--tile-content">
    {slot_above}
    <span class="bx--tile-content__below-the-fold">
      {child}
    </span>
  </div>
</div>
"""
        return self.format(template, values, context)


    def render_slot_above(self, values, context):
        """Render html of the slot.
        """
        template = """
<span data-tile-atf class="bx--tile-content__above-the-fold {class}" {props}>
  {child}
</span>
"""
        return self.format(template, values)


    def render_tmpl_icon(self, values, context):
        """Dynamically render a part of the component's template.
        """
        if 'icon' in self.slots:
            return self.format('{slot_icon}', values, context)
        return """
<svg focusable="false" preserveAspectRatio="xMidYMid meet"
    xmlns="http://www.w3.org/2000/svg" fill="currentColor" width="16"
    height="16" viewBox="0 0 16 16" aria-hidden="true">
  <path d="M8 11L3 6 3.7 5.3 8 9.6 12.3 5.3 13 6z"></path>
</svg>
"""


class TileSelectable(ChoiceFormNode):
    """Tile selectable component.
    """
    WANT_CHILDREN = True
    "Template Tag needs closing end tag."
    MODES = ('default', 'inside')
    "Available variants."

    def prepare(self, values, context):
        """Prepare values for rendering the templates.
        """
        if values['label']:
            values['props'].append(('title', values['label']))
            values['label_props'].append(('aria-label', values['label']))


    def prepare_element_props(self, props, context):
        """Prepare html attributes for rendering the form element.
        """
        props['tabindex'] = '-1'
        props['data-tile-input'] = True
        props['class'].append('bx--tile-input')


    def render_default(self, values, context):
        """Output html of the component.
        """
        template = """
{tmpl_element}
<label for="{id}" class="bx--tile bx--tile--selectable {label_class}"
    data-tile="selectable" tabindex="0" {label_props}>
  <div class="bx--tile__checkmark">
    <svg focusable="false" preserveAspectRatio="xMidYMid meet"
        xmlns="http://www.w3.org/2000/svg" fill="currentColor" width="16"
        height="16" viewBox="0 0 16 16" aria-hidden="true">
      <path d="M8,1C4.1,1,1,4.1,1,8c0,3.9,3.1,7,7,7s7-3.1,7-7C15,4.1,11.9,1,8,1z M7,11L4.3,8.3l0.9-0.8L7,9.3l4-3.9l0.9,0.8L7,11z"></path>
      <path d="M7,11L4.3,8.3l0.9-0.8L7,9.3l4-3.9l0.9,0.8L7,11z" data-icon-path="inner-path" opacity="0"></path>
    </svg>
  </div>
  <div class="bx--tile-content">
    {child}
  </div>
</label>
"""
        return self.format(template, values, context)


    def render_inside(self, values, context):
        """Output html of the component.
        """
        template = """
<label class="bx--tile bx--tile--selectable {label_class}"
    data-tile="selectable" tabindex="0" {label_props}>
  {tmpl_element}
  <div class="bx--tile__checkmark">
    <svg width="16" height="16" viewBox="0 0 16 16">
      <path d="M8 16A8 8 0 1 1 8 0a8 8 0 0 1 0 16zm3.646-10.854L6.75 10.043 4.354 7.646l-.708.708 3.104 3.103 5.604-5.603-.708-.708z"
        fill-rule="evenodd" />
    </svg>
  </div>
  <div class="bx--tile-content">
    {child}
  </div>
</label>
"""
        return self.format(template, values, context)


class TileGrid(Node):
    """Tile grid.
    """
    WANT_CHILDREN = True
    "Template Tag needs closing end tag."

    def render_default(self, values, context):
        """Output html of the component.
        """
        template = """
<div class="bx--grid {class}" {props}>
  <div class="bx--tile-container" style="width: 100%">
    {child}
  </div>
</div>
"""
        return self.format(template, values)


components = {
    'Tile': Tile,
    'TileSelect': TileSelectable,
    'TileGrid': TileGrid,
}
